# Copyright (c) 2016 QNAP Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for QNAP Storage.
This driver supports QNAP Storage for iSCSI.
"""
import base64
from collections import OrderedDict
import eventlet
import functools
import re
import threading
import time

from defusedxml import cElementTree as ET
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import units
import requests
import six
from six.moves import urllib

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume import configuration
from cinder.volume.drivers.san import san

LOG = logging.getLogger(__name__)

qnap_opts = [
    cfg.URIOpt('qnap_management_url',
               help='The URL to management QNAP Storage. '
                    'Driver does not support IPv6 address in URL.'),
    cfg.StrOpt('qnap_poolname',
               help='The pool name in the QNAP Storage'),
    cfg.StrOpt('qnap_storage_protocol',
               default='iscsi',
               help='Communication protocol to access QNAP storage')
]

CONF = cfg.CONF
CONF.register_opts(qnap_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class QnapISCSIDriver(san.SanISCSIDriver):
    """QNAP iSCSI based cinder driver


    .. code-block:: none

      Version History:
        1.0.0:
              Initial driver (Only iSCSI).
        1.2.001:
              Add supports for Thin Provisioning, SSD Cache, Deduplication,
              Compression and CHAP.
        1.2.002:
              Add support for QES fw 2.0.0.
        1.2.003:
              Add support for QES fw 2.1.0.
        1.2.004:
              Add support for QES fw on TDS series NAS model.
        1.2.005:
              Add support for QTS fw 4.4.0.

    NOTE: Set driver_ssl_cert_verify as True under backend section to
          enable SSL verification.
    """

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "QNAP_CI"

    VERSION = '1.2.005'

    TIME_INTERVAL = 3

    def __init__(self, *args, **kwargs):
        """Initialize QnapISCSIDriver."""
        super(QnapISCSIDriver, self).__init__(*args, **kwargs)
        self.api_executor = None
        self.group_stats = {}
        self.configuration.append_config_values(qnap_opts)
        self.cache_time = 0
        self.initiator = ''
        self.iscsi_port = ''
        self.target_index = ''
        self.target_iqn = ''
        self.target_iqns = []
        self.nasInfoCache = {}

    @staticmethod
    def get_driver_options():
        return qnap_opts

    def _check_config(self):
        """Ensure that the flags we care about are set."""
        LOG.debug('in _check_config')
        required_config = ['qnap_management_url',
                           'san_login',
                           'san_password',
                           'qnap_poolname',
                           'qnap_storage_protocol']

        for attr in required_config:
            if not getattr(self.configuration, attr, None):
                raise exception.InvalidInput(
                    reason=_('%s is not set.') % attr)

        if not self.configuration.use_chap_auth:
            self.configuration.chap_username = ''
            self.configuration.chap_password = ''
        else:
            if not str.isalnum(self.configuration.chap_username):
                # invalid chap_username
                LOG.error('Username must be single-byte alphabet or number.')
                raise exception.InvalidInput(
                    reason=_('Username must be single-byte '
                             'alphabet or number.'))
            if not 12 <= len(self.configuration.chap_password) <= 16:
                # invalid chap_password
                LOG.error('Password must contain 12-16 characters.')
                raise exception.InvalidInput(
                    reason=_('Password must contain 12-16 characters.'))

    def do_setup(self, context):
        """Setup the QNAP Cinder volume driver."""
        self._check_config()
        self.ctxt = context
        LOG.debug('context: %s', context)

        # Setup API Executor
        try:
            self.api_executor = self.create_api_executor()
        except Exception:
            LOG.error('Failed to create HTTP client. '
                      'Check ip, port, username, password'
                      ' and make sure the array version is compatible')
            msg = _('Failed to create HTTP client.')
            raise exception.VolumeDriverException(message=msg)

    def check_for_setup_error(self):
        """Check the status of setup."""
        pass

    def create_api_executor(self):
        """Create api executor by nas model."""
        self.api_executor = QnapAPIExecutor(
            username=self.configuration.san_login,
            password=self.configuration.san_password,
            management_url=self.configuration.qnap_management_url,
            verify_ssl=self.configuration.driver_ssl_cert_verify)

        nas_model_name, internal_model_name, fw_version = (
            self.api_executor.get_basic_info(
                self.configuration.qnap_management_url))

        if (self.configuration.qnap_management_url not in self.nasInfoCache):
            self.nasInfoCache[self.configuration.qnap_management_url] = (
                nas_model_name, internal_model_name, fw_version)

        pattern = re.compile(r"^([A-Z]+)-?[A-Z]{0,2}(\d+)\d{2}(U|[a-z]*)")
        matches = pattern.match(nas_model_name)

        if not matches:
            return None
        model_type = matches.group(1)

        ts_model_types = [
            "TS", "SS", "IS", "TVS", "TBS"
        ]
        tes_model_types = [
            "TES", "TDS"
        ]
        es_model_types = [
            "ES"
        ]
        LOG.debug('fw_version: %s', fw_version)
        if model_type in ts_model_types:
            if (fw_version >= "4.2") and (fw_version <= "4.4.9999"):
                LOG.debug('Create TS API Executor')
                # modify the pool name to pool index
                self.configuration.qnap_poolname = (
                    self._get_ts_model_pool_id(
                        self.configuration.qnap_poolname))

                return (QnapAPIExecutorTS(
                    username=self.configuration.san_login,
                    password=self.configuration.san_password,
                    management_url=self.configuration.qnap_management_url,
                    verify_ssl=self.configuration.driver_ssl_cert_verify))
        elif model_type in tes_model_types:
            if 'TS' in internal_model_name:
                if (fw_version >= "4.2") and (fw_version <= "4.4.9999"):
                    LOG.debug('Create TS API Executor')
                    # modify the pool name to poole index
                    self.configuration.qnap_poolname = (
                        self._get_ts_model_pool_id(
                            self.configuration.qnap_poolname))
                    return (QnapAPIExecutorTS(
                        username=self.configuration.san_login,
                        password=self.configuration.san_password,
                        management_url=self.configuration.qnap_management_url,
                        verify_ssl=self.configuration.driver_ssl_cert_verify))
            elif "1.1.2" <= fw_version <= "2.1.9999":
                LOG.debug('Create TES API Executor')
                return (QnapAPIExecutorTES(
                    username=self.configuration.san_login,
                    password=self.configuration.san_password,
                    management_url=self.configuration.qnap_management_url,
                    verify_ssl=self.configuration.driver_ssl_cert_verify))
        elif model_type in es_model_types:
            if "1.1.2" <= fw_version <= "2.1.9999":
                LOG.debug('Create ES API Executor')
                return (QnapAPIExecutor(
                    username=self.configuration.san_login,
                    password=self.configuration.san_password,
                    management_url=self.configuration.qnap_management_url,
                    verify_ssl=self.configuration.driver_ssl_cert_verify))

        msg = _('Model not support')
        raise exception.VolumeDriverException(message=msg)

    def _get_ts_model_pool_id(self, pool_name):
        """Modify the pool name to poole index."""
        pattern = re.compile(r"^(\d+)+|^Storage Pool (\d+)+")
        matches = pattern.match(pool_name)
        if matches.group(1):
            return matches.group(1)
        else:
            return matches.group(2)

    def _gen_random_name(self):
        return "cinder-{0}".format(timeutils.
                                   utcnow().
                                   strftime('%Y%m%d%H%M%S%f'))

    def _get_volume_metadata(self, volume):
        volume_metadata = {}
        if 'volume_metadata' in volume:
            for metadata in volume['volume_metadata']:
                volume_metadata[metadata['key']] = metadata['value']
        return volume_metadata

    def _gen_lun_name(self):
        create_lun_name = ''
        while True:
            create_lun_name = self._gen_random_name()
            # If lun name with the name exists, need to change to
            # a different name
            created_lun = self.api_executor.get_lun_info(
                LUNName=create_lun_name)
            if created_lun is None:
                break
        return create_lun_name

    def _parse_boolean_extra_spec(self, extra_spec_value):
        """Parse boolean value from extra spec.

        Parse extra spec values of the form '<is> True' , '<is> False',
        'True' and 'False'.
        """

        if not isinstance(extra_spec_value, six.string_types):
            extra_spec_value = six.text_type(extra_spec_value)

        match = re.match(r'^<is>\s*(?P<value>True|False)$',
                         extra_spec_value.strip(),
                         re.IGNORECASE)
        if match:
            extra_spec_value = match.group('value')
        return strutils.bool_from_string(extra_spec_value, strict=True)

    def create_volume(self, volume):
        """Create a new volume."""
        start_time = time.time()
        LOG.debug('in create_volume')
        LOG.debug('volume: %s', volume.__dict__)
        try:
            extra_specs = volume["volume_type"]["extra_specs"]
            LOG.debug('extra_spec: %s', extra_specs)
            qnap_thin_provision = self._parse_boolean_extra_spec(
                extra_specs.get('qnap_thin_provision', 'true'))
            qnap_compression = self._parse_boolean_extra_spec(
                extra_specs.get('qnap_compression', 'true'))
            qnap_deduplication = self._parse_boolean_extra_spec(
                extra_specs.get('qnap_deduplication', 'false'))
            qnap_ssd_cache = self._parse_boolean_extra_spec(
                extra_specs.get('qnap_ssd_cache', 'false'))
        except TypeError:
            LOG.debug('Unable to retrieve extra specs info. '
                      'Use default extra spec.')
            qnap_thin_provision = True
            qnap_compression = True
            qnap_deduplication = False
            qnap_ssd_cache = False

        LOG.debug('qnap_thin_provision: %(qnap_thin_provision)s '
                  'qnap_compression: %(qnap_compression)s '
                  'qnap_deduplication: %(qnap_deduplication)s '
                  'qnap_ssd_cache: %(qnap_ssd_cache)s',
                  {'qnap_thin_provision': qnap_thin_provision,
                   'qnap_compression': qnap_compression,
                   'qnap_deduplication': qnap_deduplication,
                   'qnap_ssd_cache': qnap_ssd_cache})

        if (qnap_deduplication and not qnap_thin_provision):
            LOG.debug('Dedupe cannot be enabled without thin_provisioning.')
            raise exception.VolumeBackendAPIException(
                data=_('Dedupe cannot be enabled without thin_provisioning.'))

        # User could create two volume with the same name on horizon.
        # Therefore, We should not use display name to create lun on nas.
        create_lun_name = self._gen_lun_name()

        create_lun_index = self.api_executor.create_lun(
            volume,
            self.configuration.qnap_poolname,
            create_lun_name,
            qnap_thin_provision,
            qnap_ssd_cache,
            qnap_compression,
            qnap_deduplication)

        max_wait_sec = 600
        try_times = 0
        lun_naa = ""
        while True:
            created_lun = self.api_executor.get_lun_info(
                LUNIndex=create_lun_index)
            if (created_lun is not None and
                    created_lun.find('LUNNAA').text is not None):
                lun_naa = created_lun.find('LUNNAA').text

            try_times = try_times + 3
            eventlet.sleep(self.TIME_INTERVAL)
            if(try_times > max_wait_sec or lun_naa != ""):
                break

        LOG.debug('LUNNAA: %s', lun_naa)
        _metadata = self._get_volume_metadata(volume)

        _metadata['LUNIndex'] = create_lun_index
        _metadata['LUNNAA'] = lun_naa
        _metadata['LunName'] = create_lun_name

        elapsed_time = time.time() - start_time
        LOG.debug('create_volume elapsed_time: %s', elapsed_time)

        LOG.debug('create_volume volid: %(volid)s, metadata: %(meta)s',
                  {'volid': volume['id'], 'meta': _metadata})

        return {'metadata': _metadata}

    @lockutils.synchronized('delete_volume', 'cinder-', True)
    def delete_volume(self, volume):
        """Delete the specified volume."""
        start_time = time.time()
        LOG.debug('volume: %s', volume.__dict__)
        lun_naa = self._get_lun_naa_from_volume_metadata(volume)
        if lun_naa == '':
            LOG.debug('Volume %s does not exist.', volume.id)
            return

        lun_index = ''
        for metadata in volume['volume_metadata']:
            if metadata['key'] == 'LUNIndex':
                lun_index = metadata['value']
                break
        LOG.debug('LUNIndex: %s', lun_index)

        internal_model_name = (self.nasInfoCache
                               [self.configuration.qnap_management_url][1])
        LOG.debug('internal_model_name: %s', internal_model_name)
        fw_version = self.nasInfoCache[self.configuration
                                           .qnap_management_url][2]
        LOG.debug('fw_version: %s', fw_version)

        if 'TS' in internal_model_name.upper():
            LOG.debug('in TS FW: get_one_lun_info')
            ret = self.api_executor.get_one_lun_info(lun_index)
            del_lun = ET.fromstring(ret['data']).find('LUNInfo').find('row')
        elif 'ES' in internal_model_name.upper():
            if fw_version >= "1.1.2" and fw_version <= "1.1.3":
                LOG.debug('in ES FW before 1.1.2/1.1.3: get_lun_info')
                del_lun = self.api_executor.get_lun_info(
                    LUNIndex=lun_index)
            elif "1.1.4" <= fw_version <= "2.1.9999":
                LOG.debug('in ES FW after 1.1.4: get_one_lun_info')
                ret = self.api_executor.get_one_lun_info(lun_index)
                del_lun = (ET.fromstring(ret['data']).find('LUNInfo')
                           .find('row'))

        if del_lun is None:
            LOG.debug('Volume %s does not exist.', lun_naa)
            return

        # if lun is mapping at target, the delete action will fail
        if del_lun.find('LUNStatus').text == '2':
            target_index = (del_lun.find('LUNTargetList')
                            .find('row').find('targetIndex').text)
            LOG.debug('target_index: %s', target_index)
            self.api_executor.disable_lun(lun_index, target_index)
            self.api_executor.unmap_lun(lun_index, target_index)

        retry_delete = False
        while True:
            retry_delete = self.api_executor.delete_lun(lun_index)
            if not retry_delete:
                break

        elapsed_time = time.time() - start_time
        LOG.debug('delete_volume elapsed_time: %s', elapsed_time)

    def _get_lun_naa_from_volume_metadata(self, volume):
        lun_naa = ''
        for metadata in volume['volume_metadata']:
            if metadata['key'] == 'LUNNAA':
                lun_naa = metadata['value']
                break
        return lun_naa

    def _extend_lun(self, volume, lun_naa):
        LOG.debug('volume: %s', volume.__dict__)
        if lun_naa == '':
            lun_naa = self._get_lun_naa_from_volume_metadata(volume)

        LOG.debug('lun_naa: %s', lun_naa)
        selected_lun = self.api_executor.get_lun_info(
            LUNNAA=lun_naa)
        lun_index = selected_lun.find('LUNIndex').text
        LOG.debug('LUNIndex: %s', lun_index)
        lun_name = selected_lun.find('LUNName').text
        LOG.debug('LUNName: %s', lun_name)
        lun_thin_allocate = selected_lun.find('LUNThinAllocate').text
        LOG.debug('LUNThinAllocate: %s', lun_thin_allocate)
        lun_path = ''
        if selected_lun.find('LUNPath') is not None:
            lun_path = selected_lun.find('LUNPath').text
            LOG.debug('LUNPath: %s', lun_path)
        lun_status = selected_lun.find('LUNStatus').text
        LOG.debug('LUNStatus: %s', lun_status)

        lun = {'LUNName': lun_name,
               'LUNCapacity': volume['size'],
               'LUNIndex': lun_index,
               'LUNThinAllocate': lun_thin_allocate,
               'LUNPath': lun_path,
               'LUNStatus': lun_status}
        self.api_executor.edit_lun(lun)

    def _create_snapshot_name(self, lun_index):
        create_snapshot_name = ''
        while True:
            # If snapshot with the name exists, need to change to
            # a different name
            create_snapshot_name = 'Q%d' % int(time.time())
            snapshot = self.api_executor.get_snapshot_info(
                lun_index=lun_index, snapshot_name=create_snapshot_name)
            if snapshot is None:
                break
        return create_snapshot_name

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the specified volume."""
        LOG.debug('Entering create_cloned_volume...')
        LOG.debug('volume: %s', volume.__dict__)
        LOG.debug('src_vref: %s', src_vref.__dict__)
        LOG.debug('volume_metadata: %s', volume['volume_metadata'])
        src_lun_naa = self._get_lun_naa_from_volume_metadata(src_vref)
        # Below is to clone a volume from a snapshot in the snapshot manager
        src_lun = self.api_executor.get_lun_info(
            LUNNAA=src_lun_naa)
        lun_index = src_lun.find('LUNIndex').text
        LOG.debug('LUNIndex: %s', lun_index)

        # User could create two snapshot with the same name on horizon.
        # Therefore, we should not use displayname to create snapshot on nas.
        create_snapshot_name = self._create_snapshot_name(lun_index)

        self.api_executor.create_snapshot_api(lun_index, create_snapshot_name)
        created_snapshot = self.api_executor.get_snapshot_info(
            lun_index=lun_index, snapshot_name=create_snapshot_name)
        snapshot_id = created_snapshot.find('snapshot_id').text
        LOG.debug('snapshot_id: %s', snapshot_id)

        # User could create two volume with the same name on horizon.
        # Therefore, We should not use displayname to create lun on nas.
        while True:
            cloned_lun_name = self._gen_random_name()
            # If lunname with the name exists, need to change to
            # a different name
            cloned_lun = self.api_executor.get_lun_info(
                LUNName=cloned_lun_name)

            if cloned_lun is None:
                break

        self.api_executor.clone_snapshot(snapshot_id, cloned_lun_name)

        max_wait_sec = 600
        try_times = 0
        lun_naa = ""
        lun_index = ""
        while True:
            created_lun = self.api_executor.get_lun_info(
                LUNName=cloned_lun_name)
            if (created_lun is not None and
                    created_lun.find('LUNNAA') is not None):
                lun_naa = created_lun.find('LUNNAA').text
                lun_index = created_lun.find('LUNIndex').text
                LOG.debug('LUNIndex: %s', lun_index)

            try_times = try_times + 3
            eventlet.sleep(self.TIME_INTERVAL)
            if(try_times > max_wait_sec or lun_naa != ""):
                break

        LOG.debug('LUNNAA: %s', lun_naa)
        if (volume['size'] > src_vref['size']):
            self._extend_lun(volume, lun_naa)
        internal_model_name = (self.nasInfoCache
                               [self.configuration.qnap_management_url][1])

        if 'TS' in internal_model_name.upper():
            LOG.debug('in TS FW: delete_snapshot_api')
            self.api_executor.delete_snapshot_api(snapshot_id)
        elif 'ES' in internal_model_name.upper():
            LOG.debug('in ES FW: do nothing')
        _metadata = self._get_volume_metadata(volume)
        _metadata['LUNIndex'] = lun_index
        _metadata['LUNNAA'] = lun_naa
        _metadata['LunName'] = cloned_lun_name
        return {'metadata': _metadata}

    def create_snapshot(self, snapshot):
        """Create a snapshot."""
        LOG.debug('snapshot: %s', snapshot.__dict__)
        LOG.debug('snapshot id: %s', snapshot['id'])

        # Below is to create snapshot in the snapshot manager
        LOG.debug('volume_metadata: %s', snapshot.volume['metadata'])
        volume_metadata = snapshot.volume['metadata']
        LOG.debug('lun_naa: %s', volume_metadata['LUNNAA'])
        lun_naa = volume_metadata['LUNNAA']
        src_lun = self.api_executor.get_lun_info(LUNNAA=lun_naa)
        lun_index = src_lun.find('LUNIndex').text
        LOG.debug('LUNIndex: %s', lun_index)

        # User could create two snapshot with the same name on horizon.
        # Therefore, We should not use displayname to create snapshot on nas.
        create_snapshot_name = self._create_snapshot_name(lun_index)
        LOG.debug('create_snapshot_name: %s', create_snapshot_name)

        self.api_executor.create_snapshot_api(lun_index, create_snapshot_name)
        max_wait_sec = 600
        try_times = 0
        snapshot_id = ""
        while True:
            created_snapshot = self.api_executor.get_snapshot_info(
                lun_index=lun_index, snapshot_name=create_snapshot_name)
            if (created_snapshot is not None and
                    created_snapshot.find('snapshot_id').text is not None):
                snapshot_id = created_snapshot.find('snapshot_id').text

            try_times = try_times + 3
            eventlet.sleep(self.TIME_INTERVAL)
            if(try_times > max_wait_sec or snapshot_id != ""):
                break

        LOG.debug('created_snapshot: %s', created_snapshot)
        LOG.debug('snapshot_id: %s', snapshot_id)

        _metadata = snapshot['metadata']
        _metadata['snapshot_id'] = snapshot_id
        _metadata['SnapshotName'] = create_snapshot_name
        return {'metadata': _metadata}

    def delete_snapshot(self, snapshot):
        """Delete a snapshot."""
        LOG.debug('snapshot: %s', snapshot.__dict__)

        # Below is to delete snapshot in the snapshot manager
        snap_metadata = snapshot['metadata']
        if 'snapshot_id' not in snap_metadata:
            return
        LOG.debug('snapshot_id: %s', snap_metadata['snapshot_id'])
        snapshot_id = snap_metadata['snapshot_id']

        self.api_executor.delete_snapshot_api(snapshot_id)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""
        LOG.debug('in create_volume_from_snapshot')
        LOG.debug('volume: %s', volume.__dict__)
        LOG.debug('snapshot: %s', snapshot.__dict__)
        # Below is to clone a volume from a snapshot in the snapshot manager
        snap_metadata = snapshot['metadata']
        if 'snapshot_id' not in snap_metadata:
            LOG.debug('Metadata of the snapshot is invalid')
            msg = _('Metadata of the snapshot is invalid')
            raise exception.VolumeDriverException(message=msg)
        LOG.debug('snapshot_id: %s', snap_metadata['snapshot_id'])
        snapshot_id = snap_metadata['snapshot_id']

        # User could create two volume with the same name on horizon.
        # Therefore, We should not use displayname to create lun on nas.
        create_lun_name = self._gen_lun_name()

        self.api_executor.clone_snapshot(
            snapshot_id, create_lun_name)

        max_wait_sec = 600
        try_times = 0
        lun_naa = ""
        lun_index = ""
        while True:
            created_lun = self.api_executor.get_lun_info(
                LUNName=create_lun_name)
            if (created_lun is not None and
                    created_lun.find('LUNNAA') is not None):
                lun_naa = created_lun.find('LUNNAA').text
                lun_index = created_lun.find('LUNIndex').text
                LOG.debug('LUNNAA: %s', lun_naa)
                LOG.debug('LUNIndex: %s', lun_index)

            try_times = try_times + 3
            eventlet.sleep(self.TIME_INTERVAL)
            if(try_times > max_wait_sec or lun_naa != ""):
                break

        if (volume['size'] > snapshot['volume_size']):
            self._extend_lun(volume, lun_naa)

        _metadata = self._get_volume_metadata(volume)
        _metadata['LUNIndex'] = lun_index
        _metadata['LUNNAA'] = lun_naa
        _metadata['LunName'] = create_lun_name
        return {'metadata': _metadata}

    def get_volume_stats(self, refresh=False):
        """Get volume stats. This is more of getting group stats."""
        LOG.debug('in get_volume_stats refresh: %s', refresh)

        if refresh:
            backend_name = (self.configuration.safe_get(
                            'volume_backend_name') or
                            self.__class__.__name__)
            LOG.debug('backend_name=%(backend_name)s',
                      {'backend_name': backend_name})

            selected_pool = self.api_executor.get_specific_poolinfo(
                self.configuration.qnap_poolname)
            capacity_bytes = int(selected_pool.find('capacity_bytes').text)
            LOG.debug('capacity_bytes: %s GB', capacity_bytes / units.Gi)
            freesize_bytes = int(selected_pool.find('freesize_bytes').text)
            LOG.debug('freesize_bytes: %s GB', freesize_bytes / units.Gi)
            provisioned_bytes = int(selected_pool.find('allocated_bytes').text)
            driver_protocol = self.configuration.qnap_storage_protocol
            LOG.debug(
                'provisioned_bytes: %s GB', provisioned_bytes / units.Gi)
            self.group_stats = {'volume_backend_name': backend_name,
                                'vendor_name': 'QNAP',
                                'driver_version': self.VERSION,
                                'storage_protocol': driver_protocol}
            # single pool now, need support multiple pools in the future
            single_pool = dict(
                pool_name=self.configuration.qnap_poolname,
                total_capacity_gb=capacity_bytes / units.Gi,
                free_capacity_gb=freesize_bytes / units.Gi,
                provisioned_capacity_gb=provisioned_bytes / units.Gi,
                reserved_percentage=self.configuration.reserved_percentage,
                QoS_support=False,
                qnap_thin_provision=["True", "False"],
                qnap_compression=["True", "False"],
                qnap_deduplication=["True", "False"],
                qnap_ssd_cache=["True", "False"])
            self.group_stats['pools'] = [single_pool]

        return self.group_stats

    def extend_volume(self, volume, new_size):
        """Extend an existing volume."""
        LOG.debug('Entering extend_volume volume=%(vol)s '
                  'new_size=%(size)s',
                  {'vol': volume['display_name'], 'size': new_size})

        volume['size'] = new_size
        self._extend_lun(volume, '')

    def _get_portal_info(self, volume, connector, lun_slot_id, lun_owner):
        """Get portal info."""
        # Cache portal info for twenty seconds
        # If connectors were the same then use the portal info which was cached
        LOG.debug('get into _get_portal_info')
        self.initiator = connector['initiator']
        ret = self.api_executor.get_iscsi_portal_info()
        root = ET.fromstring(ret['data'])
        iscsi_port = root.find('iSCSIPortal').find('servicePort').text
        LOG.debug('iscsiPort: %s', iscsi_port)
        target_iqn_prefix = root.find(
            'iSCSIPortal').find('targetIQNPrefix').text
        LOG.debug('targetIQNPrefix: %s', target_iqn_prefix)

        internal_model_name = (self.nasInfoCache
                               [self.configuration.qnap_management_url][1])
        LOG.debug('internal_model_name: %s', internal_model_name)
        fw_version = (self.nasInfoCache
                      [self.configuration.qnap_management_url][2])
        LOG.debug('fw_version: %s', fw_version)

        target_index = ''
        target_iqn = ''

        # create a new target if no target has ACL connector['initiator']
        LOG.debug('exist target_index: %s', target_index)
        if not target_index:
            target_name = self._gen_random_name()
            LOG.debug('target_name: %s', target_name)
            target_index = self.api_executor.create_target(
                target_name, lun_owner)
            LOG.debug('targetIndex: %s', target_index)

            retryCount = 0
            retrySleepTime = 2
            while retryCount <= 5:
                target_info = self.api_executor.get_target_info(target_index)
                if target_info.find('targetIQN').text is not None:
                    break
                eventlet.sleep(retrySleepTime)
                retrySleepTime = retrySleepTime + 2
                retryCount = retryCount + 1

            target_iqn = target_info.find('targetIQN').text
            LOG.debug('target_iqn: %s', target_iqn)

            # TS NAS have to remove default ACL
            default_acl = (
                target_iqn_prefix[:target_iqn_prefix.find(":") + 1])
            default_acl = default_acl + "all:iscsi.default.ffffff"
            LOG.debug('default_acl: %s', default_acl)
            self.api_executor.remove_target_init(target_iqn, default_acl)
            # add ACL
            self.api_executor.add_target_init(
                target_iqn, connector['initiator'],
                self.configuration.use_chap_auth,
                self.configuration.chap_username,
                self.configuration.chap_password)

        # Get information for multipath
        target_iqns = []
        slotid_list = []
        eth_list, slotid_list = Util.retriveFormCache(
            self.configuration.qnap_management_url,
            lambda: self.api_executor.get_ethernet_ip(type='data'),
            30)

        LOG.debug('slotid_list: %s', slotid_list)
        target_portals = []
        target_portals.append(
            self.configuration.target_ip_address + ':' + iscsi_port)
        # target_iqns.append(target_iqn)
        for index, eth in enumerate(eth_list):
            # TS NAS do not have slot_id
            if not slotid_list:
                target_iqns.append(target_iqn)
            else:
                # To support ALUA, target portal and target inq should
                # be consistent.
                # EX: 10.77.230.31:3260 at controller B and it should map
                # to the target at controller B
                target_iqns.append(
                    target_iqn[:-2] + '.' + slotid_list[index])

            if eth == self.configuration.target_ip_address:
                continue
            target_portals.append(eth + ':' + iscsi_port)

        self.iscsi_port = iscsi_port
        self.target_index = target_index
        self.target_iqn = target_iqn
        self.target_iqns = target_iqns
        self.target_portals = target_portals

        return (iscsi_port, target_index, target_iqn,
                target_iqns, target_portals)

    @lockutils.synchronized('create_export', 'cinder-', True)
    def create_export(self, context, volume, connector):
        start_time = time.time()
        LOG.debug('in create_export')
        LOG.debug('volume: %s', volume.__dict__)
        LOG.debug('connector: %s', connector)

        lun_naa = self._get_lun_naa_from_volume_metadata(volume)
        if lun_naa == '':
            msg = (_("Volume %s does not exist.") % volume.id)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        LOG.debug('volume[name]: %s', volume['name'])
        LOG.debug('volume[display_name]: %s', volume['display_name'])

        lun_index = ''
        for metadata in volume['volume_metadata']:
            if metadata['key'] == 'LUNIndex':
                lun_index = metadata['value']
                break
        LOG.debug('LUNIndex: %s', lun_index)
        internal_model_name = (self.nasInfoCache
                               [self.configuration.qnap_management_url][1])
        LOG.debug('internal_model_name: %s', internal_model_name)
        fw_version = self.nasInfoCache[self.configuration
                                           .qnap_management_url][2]
        LOG.debug('fw_version: %s', fw_version)
        if 'TS' in internal_model_name.upper():
            LOG.debug('in TS FW: get_one_lun_info')
            ret = self.api_executor.get_one_lun_info(lun_index)
            selected_lun = (ET.fromstring(ret['data']).find('LUNInfo')
                            .find('row'))
        elif 'ES' in internal_model_name.upper():
            if fw_version >= "1.1.2" and fw_version <= "1.1.3":
                LOG.debug('in ES FW before 1.1.2/1.1.3: get_lun_info')
                selected_lun = self.api_executor.get_lun_info(
                    LUNNAA=lun_naa)
            elif "1.1.4" <= fw_version <= "2.1.9999":
                LOG.debug('in ES FW after 1.1.4: get_one_lun_info')
                ret = self.api_executor.get_one_lun_info(lun_index)
                selected_lun = (ET.fromstring(ret['data']).find('LUNInfo')
                                .find('row'))

        lun_owner = ''
        lun_slot_id = ''
        if selected_lun.find('lun_owner') is not None:
            lun_owner = selected_lun.find('lun_owner').text
            LOG.debug('lun_owner: %s', lun_owner)
            lun_slot_id = '0' if (lun_owner == 'SCA') else '1'
            LOG.debug('lun_slot_id: %s', lun_slot_id)

        # LOG.debug('self.initiator: %s', self.initiator)
        LOG.debug('connector: %s', connector['initiator'])

        iscsi_port, target_index, target_iqn, target_iqns, target_portals = (
            self._get_portal_info(volume, connector, lun_slot_id, lun_owner))

        self.api_executor.map_lun(lun_index, target_index)

        max_wait_sec = 600
        try_times = 0
        LUNNumber = ""
        target_lun_id = -999
        while True:
            if 'TS' in internal_model_name.upper():
                LOG.debug('in TS FW: get_one_lun_info')
                ret = self.api_executor.get_one_lun_info(lun_index)
                root = ET.fromstring(ret['data'])
                target_lun_id = int(root.find('LUNInfo').find('row')
                                    .find('LUNTargetList').find('row')
                                    .find('LUNNumber').text)

                try_times = try_times + 3
                eventlet.sleep(self.TIME_INTERVAL)
                if(try_times > max_wait_sec or target_lun_id != -999):
                    break

            elif 'ES' in internal_model_name.upper():
                if fw_version >= "1.1.2" and fw_version <= "1.1.3":
                    LOG.debug('in ES FW before 1.1.2/1.1.3: get_lun_info')
                    root = self.api_executor.get_lun_info(LUNNAA=lun_naa)
                    if len(list(root.find('LUNTargetList'))) != 0:
                        LUNNumber = root.find('LUNTargetList').find(
                            'row').find('LUNNumber').text
                    target_lun_id = int(LUNNumber)

                    try_times = try_times + 3
                    eventlet.sleep(self.TIME_INTERVAL)
                    if(try_times > max_wait_sec or LUNNumber != ""):
                        break
                elif "1.1.4" <= fw_version <= "2.1.9999":
                    LOG.debug('in ES FW after 1.1.4: get_one_lun_info')
                    ret = self.api_executor.get_one_lun_info(lun_index)
                    root = ET.fromstring(ret['data'])
                    target_lun_id = int(root.find('LUNInfo')
                                        .find('row').find('LUNTargetList')
                                        .find('row').find('LUNNumber').text)

                    try_times = try_times + 3
                    eventlet.sleep(self.TIME_INTERVAL)
                    if(try_times > max_wait_sec or target_lun_id != -999):
                        break
                else:
                    break
            else:
                break

        properties = {}
        properties['target_discovered'] = False
        properties['target_portal'] = (self.configuration.target_ip_address +
                                       ':' + iscsi_port)
        properties['target_iqn'] = target_iqn
        LOG.debug('properties[target_iqn]: %s', properties['target_iqn'])

        LOG.debug('target_lun_id: %s', target_lun_id)
        properties['target_lun'] = target_lun_id
        properties['volume_id'] = volume['id']  # used by xen currently

        multipath = connector.get('multipath', False)
        if multipath:
            """Below are settings for multipath"""
            properties['target_portals'] = target_portals
            properties['target_iqns'] = target_iqns
            properties['target_luns'] = (
                [target_lun_id] * len(target_portals))
            LOG.debug('properties: %s', properties)

        provider_location = '%(host)s:%(port)s,1 %(name)s %(tgt_lun)s' % {
            'host': self.configuration.target_ip_address,
            'port': iscsi_port,
            'name': target_iqn,
            'tgt_lun': target_lun_id,
        }

        if self.configuration.use_chap_auth:
            provider_auth = 'CHAP %s %s' % (self.configuration.chap_username,
                                            self.configuration.chap_password)
        else:
            provider_auth = None

        elapsed_time = time.time() - start_time
        LOG.debug('create_export elapsed_time: %s', elapsed_time)

        LOG.debug('create_export volid: %(volid)s, provider_location: %(loc)s',
                  {'volid': volume['id'], 'loc': provider_location})

        return (
            {'provider_location': provider_location,
             'provider_auth': provider_auth})

    def initialize_connection(self, volume, connector):
        start_time = time.time()
        LOG.debug('in initialize_connection')

        if not volume['provider_location']:
            err = _("Param volume['provider_location'] is invalid.")
            raise exception.InvalidParameterValue(err=err)

        result = volume['provider_location'].split(' ')
        if len(result) < 2:
            raise exception.InvalidInput(reason=volume['provider_location'])

        data = result[0].split(',')
        if len(data) < 2:
            raise exception.InvalidInput(reason=volume['provider_location'])

        iqn = result[1]
        LOG.debug('iqn: %s', iqn)
        target_lun_id = int(result[2], 10)
        LOG.debug('target_lun_id: %d', target_lun_id)

        properties = {}
        properties['target_discovered'] = False
        properties['target_portal'] = (self.configuration.target_ip_address +
                                       ':' + self.iscsi_port)
        properties['target_iqn'] = iqn
        properties['target_lun'] = target_lun_id
        properties['volume_id'] = volume['id']  # used by xen currently

        if self.configuration.use_chap_auth:
            properties['auth_method'] = 'CHAP'
            properties['auth_username'] = self.configuration.chap_username
            properties['auth_password'] = self.configuration.chap_password

        elapsed_time = time.time() - start_time
        LOG.debug('initialize_connection elapsed_time: %s', elapsed_time)

        LOG.debug('initialize_connection volid:'
                  ' %(volid)s, properties: %(prop)s',
                  {'volid': volume['id'], 'prop': properties})

        return {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

    def enum(self, *sequential, **named):
        """Enum method."""
        enums = dict(zip(sequential, range(len(sequential))), **named)
        return type('Enum', (), enums)

    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to unattach a volume from an instance."""

        start_time = time.time()
        LOG.debug('in terminate_connection')
        LOG.debug('volume: %s', volume.__dict__)
        LOG.debug('connector: %s', connector)

        # get lun index
        lun_naa = self._get_lun_naa_from_volume_metadata(volume)
        LOG.debug('lun_naa: %s', lun_naa)
        lun_index = ''
        for metadata in volume['volume_metadata']:
            if metadata['key'] == 'LUNIndex':
                lun_index = metadata['value']
                break
        LOG.debug('LUNIndex: %s', lun_index)

        internal_model_name = (self.nasInfoCache
                               [self.configuration.qnap_management_url][1])
        LOG.debug('internal_model_name: %s', internal_model_name)
        fw_version = self.nasInfoCache[self.configuration
                                           .qnap_management_url][2]
        LOG.debug('fw_version: %s', fw_version)

        if 'TS' in internal_model_name.upper():
            LOG.debug('in TS FW: get_one_lun_info')
            ret = self.api_executor.get_one_lun_info(lun_index)
            selected_lun = (ET.fromstring(ret['data']).find('LUNInfo')
                            .find('row'))
        elif 'ES' in internal_model_name.upper():
            if fw_version >= "1.1.2" and fw_version <= "1.1.3":
                LOG.debug('in ES FW before 1.1.2/1.1.3: get_lun_info')
                selected_lun = self.api_executor.get_lun_info(
                    LUNIndex=lun_index)
            elif "1.1.4" <= fw_version <= "2.1.9999":
                LOG.debug('in ES FW after 1.1.4: get_one_lun_info')
                ret = self.api_executor.get_one_lun_info(lun_index)
                selected_lun = (ET.fromstring(ret['data']).find('LUNInfo')
                                .find('row'))

        lun_status = self.enum('createing', 'unmapped', 'mapped')

        LOG.debug('LUNStatus: %s', selected_lun.find('LUNStatus').text)
        LOG.debug('lun_status.mapped: %s', six.text_type(lun_status.mapped))
        # lun does not map to any target
        if (selected_lun.find('LUNStatus').text) != (
                six.text_type(lun_status.mapped)):
            return

        target_index = (selected_lun.find('LUNTargetList')
                        .find('row').find('targetIndex').text)
        LOG.debug('target_index: %s', target_index)

        start_time1 = time.time()
        self.api_executor.disable_lun(lun_index, target_index)
        elapsed_time1 = time.time() - start_time1
        LOG.debug('terminate_connection disable_lun elapsed_time : %s',
                  elapsed_time1)

        start_time2 = time.time()
        self.api_executor.unmap_lun(lun_index, target_index)
        elapsed_time2 = time.time() - start_time2
        LOG.debug('terminate_connection unmap_lun elapsed_time : %s',
                  elapsed_time2)

        elapsed_time = time.time() - start_time

        LOG.debug('terminate_connection elapsed_time : %s', elapsed_time)
        self.api_executor.delete_target(target_index)

    def update_migrated_volume(
            self, context, volume, new_volume, original_volume_status):
        """Return model update for migrated volume."""
        LOG.debug('volume: %s', volume.__dict__)
        LOG.debug('new_volume: %s', new_volume.__dict__)
        LOG.debug('original_volume_status: %s', original_volume_status)

        _metadata = self._get_volume_metadata(new_volume)

        # metadata will not be swap after migration with liberty version
        # and the metadata of new volume is different with the metadata
        # of original volume. Therefore, we need to update the migrated volume.
        if not hasattr(new_volume, '_orig_metadata'):
            model_update = {'metadata': _metadata}
            return model_update

    @utils.synchronized('_attach_volume')
    def _detach_volume(self, context, attach_info, volume, properties,
                       force=False, remote=False, ignore_errors=False):
        super(QnapISCSIDriver, self)._detach_volume(
            context, attach_info,
            volume, properties,
            force=force, remote=remote,
            ignore_errors=ignore_errors
        )

    @utils.synchronized('_attach_volume')
    def _attach_volume(self, context, volume, properties, remote=False):
        return super(QnapISCSIDriver, self)._attach_volume(context, volume,
                                                           properties, remote)


def _connection_checker(func):
    """Decorator to check session has expired or not."""
    @functools.wraps(func)
    def inner_connection_checker(self, *args, **kwargs):
        LOG.debug('in _connection_checker')
        for attempts in range(5):
            try:
                return func(self, *args, **kwargs)
            except exception.VolumeBackendAPIException as e:
                pattern = re.compile(
                    r".*Session id expired$")
                matches = pattern.match(six.text_type(e))
                if matches:
                    if attempts < 4:
                        LOG.debug('Session might have expired.'
                                  ' Trying to relogin')
                        self._login()
                        continue

                LOG.error('Re-throwing Exception %s', e)
                raise
    return inner_connection_checker


class QnapAPIExecutor(object):
    """Makes QNAP API calls for ES NAS."""
    es_create_lun_lock = threading.Lock()
    es_delete_lun_lock = threading.Lock()
    es_lun_locks = {}

    def __init__(self, *args, **kwargs):
        """Init function."""
        self.sid = None
        self.username = kwargs['username']
        self.password = kwargs['password']
        self.ip, self.port, self.ssl = (
            self._parse_management_url(kwargs['management_url']))
        self.verify_ssl = kwargs['verify_ssl']
        self._login()

    def _parse_management_url(self, management_url):
        # NOTE(Ibad): This parser isn't compatible with IPv6 address.
        # Typical IPv6 address will have : as delimiters and
        # URL is represented as https://[3ffe:2a00:100:7031::1]:8080
        # since the regular expression below uses : to identify ip and port
        # it won't work with IPv6 address.
        pattern = re.compile(r"(http|https)\:\/\/(\S+)\:(\d+)")
        matches = pattern.match(management_url)
        if matches.group(1) == 'http':
            management_ssl = False
        else:
            management_ssl = True
        management_ip = matches.group(2)
        management_port = matches.group(3)
        return management_ip, management_port, management_ssl

    def get_basic_info(self, management_url):
        """Get the basic information of NAS."""
        management_ip, management_port, management_ssl = (
            self._parse_management_url(management_url))

        response = self._get_response(management_ip, management_port,
                                      management_ssl, '/cgi-bin/authLogin.cgi')
        data = response.text

        root = ET.fromstring(data)

        nas_model_name = root.find('model/displayModelName').text
        internal_model_name = root.find('model/internalModelName').text
        fw_version = root.find('firmware/version').text

        return nas_model_name, internal_model_name, fw_version

    def _get_response(self, host_ip, host_port, use_ssl, action, body=None):
        """"Execute http request and return response."""
        method = 'GET'
        headers = None
        protocol = 'https' if use_ssl else 'http'
        verify = self.verify_ssl if use_ssl else False
        # NOTE(ibad): URL formed here isn't IPv6 compatible
        # we should surround host ip with [] when IPv6 is supported
        # so the final URL can be like https://[3ffe:2a00:100:7031::1]:8080
        url = '%s://%s:%s%s' % (protocol, host_ip, host_port, action)

        if body:
            method = 'POST'
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'charset': 'utf-8'
            }

        response = requests.request(method, url, data=body, headers=headers,
                                    verify=verify)

        return response

    def _execute_and_get_response_details(self, nas_ip, url, post_parm=None):
        """Will prepare response after executing an http request."""
        LOG.debug('_execute_and_get_response_details url: %s', url)
        LOG.debug('_execute_and_get_response_details post_parm: %s', post_parm)

        res_details = {}

        # Make the connection
        start_time2 = time.time()
        response = self._get_response(
            nas_ip, self.port, self.ssl, url, post_parm)
        elapsed_time2 = time.time() - start_time2
        LOG.debug('request elapsed_time: %s', elapsed_time2)

        # Read the response
        data = response.text
        LOG.debug('response status: %s', response.status_code)

        # Extract http error msg if any
        error_details = None
        res_details['data'] = data
        res_details['error'] = error_details
        res_details['http_status'] = response.status_code

        return res_details

    def execute_login(self):
        """Login and return sid."""
        params = OrderedDict()
        params['pwd'] = base64.b64encode(self.password.encode("utf-8"))
        params['serviceKey'] = '1'
        params['user'] = self.username

        sanitized_params = OrderedDict()

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/authLogin.cgi?')

        res_details = self._execute_and_get_response_details(
            self.ip, url, sanitized_params)
        root = ET.fromstring(res_details['data'])
        LOG.debug('execute_login data: %s', res_details['data'])
        session_id = root.find('authSid').text
        LOG.debug('execute_login session_id: %s', session_id)
        return session_id

    def _login(self):
        """Execute Https Login API."""
        self.sid = self.execute_login()

    def _get_res_details(self, url, **kwargs):
        sanitized_params = OrderedDict()

        # Sort the dict of parameters
        params = utils.create_ordereddict(kwargs)

        for key, value in params.items():
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        encoded_params = urllib.parse.urlencode(sanitized_params)
        url = url + encoded_params

        res_details = self._execute_and_get_response_details(self.ip, url)

        return res_details

    @_connection_checker
    def create_lun(self, volume, pool_name, create_lun_name, reserve,
                   ssd_cache, compress, dedup):
        """Create lun."""
        self.es_create_lun_lock.acquire()

        lun_thin_allocate = ''
        if reserve:
            lun_thin_allocate = '1'
        else:
            lun_thin_allocate = '0'

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_lun_setting.cgi?',
                func='add_lun',
                FileIO='no',
                LUNThinAllocate=lun_thin_allocate,
                LUNName=create_lun_name,
                LUNPath=create_lun_name,
                poolID=pool_name,
                lv_ifssd='yes' if ssd_cache else 'no',
                compression='1' if compress else '0',
                dedup='sha256' if dedup else 'off',
                LUNCapacity=volume['size'],
                lv_threshold='80',
                sid=self.sid)
        finally:
            self.es_create_lun_lock.release()

        root = ET.fromstring(res_details['data'])

        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Create volume %s failed') % volume['display_name'])

        return root.find('result').text

    @_connection_checker
    def delete_lun(self, vol_id, *args, **kwargs):
        """Execute delete lun API."""

        self.es_delete_lun_lock.acquire()

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_lun_setting.cgi?',
                func='remove_lun',
                run_background='1',
                ha_sync='1',
                LUNIndex=vol_id,
                sid=self.sid)
        finally:
            self.es_delete_lun_lock.release()

        data_set_is_busy = "-205041"
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        # dataset is busy, retry to delete
        if root.find('result').text == data_set_is_busy:
            return True
        if root.find('result').text < '0':
            msg = (_('Volume %s delete failed') % vol_id)
            raise exception.VolumeBackendAPIException(data=msg)

        return False

    @_connection_checker
    def get_specific_poolinfo(self, pool_id):
        """Execute get specific poolinfo API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/disk_manage.cgi?',
            store='poolInfo',
            func='extra_get',
            poolID=pool_id,
            Pool_Info='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('get_specific_poolinfo failed'))

        pool_list = root.find('Pool_Index')
        pool_info_tree = pool_list.findall('row')
        for pool in pool_info_tree:
            if pool_id == pool.find('poolID').text:
                return pool

    @_connection_checker
    def create_target(self, target_name, controller_name):
        """Create target on nas and return target index."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='add_target',
            targetName=target_name,
            targetAlias=target_name,
            bTargetDataDigest='0',
            bTargetHeaderDigest='0',
            bTargetClusterEnable='1',
            controller_name=controller_name,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Create target failed'))

        root = ET.fromstring(res_details['data'])
        targetIndex = root.find('result').text
        return targetIndex

    @_connection_checker
    def delete_target(self, target_index):
        """Delete target on nas."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='remove_target',
            targetIndex=target_index,
            sid=self.sid)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text != '0':
            raise exception.VolumeBackendAPIException(
                data=_('Delete target failed'))

    @_connection_checker
    def add_target_init(self, target_iqn, init_iqn, use_chap_auth,
                        chap_username, chap_password):
        """Add target acl."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='add_init',
            targetIQN=target_iqn,
            initiatorIQN=init_iqn,
            initiatorAlias=init_iqn,
            bCHAPEnable='1' if use_chap_auth else '0',
            CHAPUserName=chap_username,
            CHAPPasswd=chap_password,
            bMutualCHAPEnable='0',
            mutualCHAPUserName='',
            mutualCHAPPasswd='',
            ha_sync='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Add target acl failed'))

    def remove_target_init(self, target_iqn, init_iqn):
        """Remote target acl."""
        pass

    @_connection_checker
    def map_lun(self, lun_index, target_index):
        """Map lun to sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='add_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                sid=self.sid)
        finally:
            pass

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                "Map lun %(lun_index)s to target %(target_index)s failed") %
                {'lun_index': six.text_type(lun_index),
                 'target_index': six.text_type(target_index)})

    @_connection_checker
    def disable_lun(self, lun_index, target_index):
        """Disable lun from sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='edit_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                LUNEnable=0,
                sid=self.sid)
        finally:
            pass

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                'Disable lun %(lun_index)s from target %(target_index)s failed'
            ) % {'lun_index': lun_index, 'target_index': target_index})

    @_connection_checker
    def unmap_lun(self, lun_index, target_index):
        """Unmap lun from sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='remove_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                sid=self.sid)
        finally:
            pass

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                'Unmap lun %(lun_index)s from target %(target_index)s failed')
                % {'lun_index': lun_index, 'target_index': target_index})

    @_connection_checker
    def get_iscsi_portal_info(self):
        """Get iscsi portal info."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            iSCSI_portal='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        else:
            return res_details

    @_connection_checker
    def get_lun_info(self, **kwargs):
        """Execute get_lun_info API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            lunList='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))

        if (('LUNIndex' in kwargs) or ('LUNName' in kwargs) or
                ('LUNNAA' in kwargs)):

            lun_list = root.find('iSCSILUNList')
            lun_info_tree = lun_list.findall('LUNInfo')
            for lun in lun_info_tree:
                if ('LUNIndex' in kwargs):
                    if (kwargs['LUNIndex'] == lun.find('LUNIndex').text):
                        return lun
                elif ('LUNName' in kwargs):
                    if (kwargs['LUNName'] == lun.find('LUNName').text):
                        return lun
                elif ('LUNNAA' in kwargs):
                    if (kwargs['LUNNAA'] == lun.find('LUNNAA').text):
                        return lun

        return None

    @_connection_checker
    def get_one_lun_info(self, lunID):
        """Execute get_one_lun_info API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            lun_info='1',
            lunID=lunID,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        else:
            return res_details

    @_connection_checker
    def get_snapshot_info(self, **kwargs):
        """Execute get_snapshot_info API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/snapshot.cgi?',
            func='extra_get',
            LUNIndex=kwargs['lun_index'],
            snapshot_list='1',
            snap_start='0',
            snap_count='100',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Unexpected response from QNAP API'))

        snapshot_list = root.find('SnapshotList')
        if snapshot_list is None:
            return None
        snapshot_tree = snapshot_list.findall('row')
        for snapshot in snapshot_tree:
            if (kwargs['snapshot_name'] ==
                    snapshot.find('snapshot_name').text):
                return snapshot

        return None

    @_connection_checker
    def create_snapshot_api(self, lun_id, snapshot_name):
        """Execute CGI to create snapshot from source lun."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/snapshot.cgi?',
            func='create_snapshot',
            lunID=lun_id,
            snapshot_name=snapshot_name,
            expire_min='0',
            vital='1',
            snapshot_type='0',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('create snapshot failed'))

    @_connection_checker
    def delete_snapshot_api(self, snapshot_id):
        """Execute CGI to delete snapshot by snapshot id."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/snapshot.cgi?',
            func='del_snapshots',
            snapshotID=snapshot_id,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        # snapshot not exist
        if root.find('result').text == '-206021':
            return
        # lun not exist
        if root.find('result').text == '-200005':
            return
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('delete snapshot %s failed') % snapshot_id)

    @_connection_checker
    def clone_snapshot(self, snapshot_id, new_lunname):
        """Execute CGI to clone snapshot as unmap lun."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/snapshot.cgi?',
            func='clone_qsnapshot',
            by_lun='1',
            snapshotID=snapshot_id,
            new_name=new_lunname,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                'Clone lun %(lunname)s from snapshot %(snapshot_id)s failed'
            ) % {'lunname': new_lunname, 'snapshot_id': snapshot_id})

    @_connection_checker
    def edit_lun(self, lun):
        """Extend lun."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_lun_setting.cgi?',
            func='edit_lun',
            LUNName=lun['LUNName'],
            LUNCapacity=lun['LUNCapacity'],
            LUNIndex=lun['LUNIndex'],
            LUNThinAllocate=lun['LUNThinAllocate'],
            LUNPath=lun['LUNPath'],
            LUNStatus=lun['LUNStatus'],
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Extend lun %s failed') % lun['LUNIndex'])

    @_connection_checker
    def get_all_iscsi_portal_setting(self):
        """Execute get_all_iscsi_portal_setting API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='get_all',
            sid=self.sid)

        return res_details

    @_connection_checker
    def get_ethernet_ip(self, **kwargs):
        """Execute get_ethernet_ip API."""
        res_details = self._get_res_details(
            '/cgi-bin/sys/sysRequest.cgi?',
            subfunc='net_setting',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))

        if ('type' in kwargs):
            return_ip = []
            return_slot_id = []
            ip_list = root.find('func').find('ownContent')
            ip_list_tree = ip_list.findall('IPInfo')
            for IP in ip_list_tree:
                ipv4 = (IP.find('IP').find('IP1').text + '.' +
                        IP.find('IP').find('IP2').text + '.' +
                        IP.find('IP').find('IP3').text + '.' +
                        IP.find('IP').find('IP4').text)
                if ((kwargs['type'] == 'data') and
                    (IP.find('isManagePort').text != '1') and
                        (IP.find('status').text == '1')):
                    return_slot_id.append(IP.find('interfaceSlotid').text)
                    return_ip.append(ipv4)
                elif ((kwargs['type'] == 'manage') and
                      (IP.find('isManagePort').text == '1') and
                      (IP.find('status').text == '1')):
                    return_ip.append(ipv4)
                elif ((kwargs['type'] == 'all') and
                      (IP.find('status').text == '1')):
                    return_ip.append(ipv4)

        return return_ip, return_slot_id

    @_connection_checker
    def get_target_info(self, target_index):
        """Get target info."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            targetInfo=1,
            targetIndex=target_index,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        LOG.debug('ES get_target_info.authPassed: (%s)',
                  root.find('authPassed').text)
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Get target info failed'))

        target_list = root.find('targetInfo')
        target_tree = target_list.findall('row')
        for target in target_tree:
            if target_index == target.find('targetIndex').text:
                return target

    @_connection_checker
    def get_target_info_by_initiator(self, initiatorIQN):
        """Get target info by initiatorIQN."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            initiatorIQN=initiatorIQN,
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            return "", ""

        target = root.find('targetACL').find('row')
        targetIndex = target.find('targetIndex').text
        targetIQN = target.find('targetIQN').text

        return targetIndex, targetIQN


class QnapAPIExecutorTS(QnapAPIExecutor):
    """Makes QNAP API calls for TS NAS."""
    create_lun_lock = threading.Lock()
    delete_lun_lock = threading.Lock()
    lun_locks = {}

    @_connection_checker
    def create_lun(self, volume, pool_name, create_lun_name, reserve,
                   ssd_cache, compress, dedup):
        """Create lun."""
        self.create_lun_lock.acquire()

        lun_thin_allocate = ''
        if reserve:
            lun_thin_allocate = '1'
        else:
            lun_thin_allocate = '0'

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_lun_setting.cgi?',
                func='add_lun',
                FileIO='no',
                LUNThinAllocate=lun_thin_allocate,
                LUNName=create_lun_name,
                LUNPath=create_lun_name,
                poolID=pool_name,
                lv_ifssd='yes' if ssd_cache else 'no',
                LUNCapacity=volume['size'],
                LUNSectorSize='512',
                lv_threshold='80',
                sid=self.sid)
        finally:
            self.create_lun_lock.release()

        root = ET.fromstring(res_details['data'])

        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Create volume %s failed') % volume['display_name'])

        return root.find('result').text

    @_connection_checker
    def delete_lun(self, vol_id, *args, **kwargs):
        """Execute delete lun API."""
        self.delete_lun_lock.acquire()

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_lun_setting.cgi?',
                func='remove_lun',
                run_background='1',
                ha_sync='1',
                LUNIndex=vol_id,
                sid=self.sid)
        finally:
            self.delete_lun_lock.release()

        data_set_is_busy = "-205041"
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        # dataset is busy, retry to delete
        if root.find('result').text == data_set_is_busy:
            return True
        if root.find('result').text < '0':
            msg = (_('Volume %s delete failed') % vol_id)
            raise exception.VolumeBackendAPIException(data=msg)

        return False

    @lockutils.synchronized('map_unmap_lun_ts')
    @_connection_checker
    def map_lun(self, lun_index, target_index):
        """Map lun to sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='add_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                sid=self.sid)
        finally:
            pass

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                "Map lun %(lun_index)s to target %(target_index)s failed") %
                {'lun_index': six.text_type(lun_index),
                 'target_index': six.text_type(target_index)})

        return root.find('result').text

    @_connection_checker
    def disable_lun(self, lun_index, target_index):
        """Disable lun from sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='edit_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                LUNEnable=0,
                sid=self.sid)
        finally:
            pass

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                'Disable lun %(lun_index)s from target %(target_index)s failed'
            ) % {'lun_index': lun_index, 'target_index': target_index})

    @_connection_checker
    def unmap_lun(self, lun_index, target_index):
        """Unmap lun from sepecific target."""

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_target_setting.cgi?',
                func='remove_lun',
                LUNIndex=lun_index,
                targetIndex=target_index,
                sid=self.sid)
        finally:
            pass
#           self.lun_locks[lun_index].release()

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(data=_(
                'Unmap lun %(lun_index)s from target %(target_index)s failed')
                % {'lun_index': lun_index, 'target_index': target_index})

    @_connection_checker
    def remove_target_init(self, target_iqn, init_iqn):
        """Remove target acl."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='remove_init',
            targetIQN=target_iqn,
            initiatorIQN=init_iqn,
            ha_sync='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Remove target acl failed'))

    @_connection_checker
    def get_target_info(self, target_index):
        """Get nas target info."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_portal_setting.cgi?',
            func='extra_get',
            targetInfo=1,
            targetIndex=target_index,
            ha_sync='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        LOG.debug('TS get_target_info.authPassed: (%s)',
                  root.find('authPassed').text)
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Get target info failed'))

        target_list = root.find('targetInfo')
        target_tree = target_list.findall('row')
        for target in target_tree:
            if target_index == target.find('targetIndex').text:
                return target

    @_connection_checker
    def get_ethernet_ip(self, **kwargs):
        """Execute get_ethernet_ip API."""
        res_details = self._get_res_details(
            '/cgi-bin/sys/sysRequest.cgi?',
            subfunc='net_setting',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))

        if ('type' in kwargs):
            return_ip = []
            ip_list = root.find('func').find('ownContent')
            ip_list_tree = ip_list.findall('IPInfo')
            for IP in ip_list_tree:
                ipv4 = (IP.find('IP').find('IP1').text + '.' +
                        IP.find('IP').find('IP2').text + '.' +
                        IP.find('IP').find('IP3').text + '.' +
                        IP.find('IP').find('IP4').text)
                if (IP.find('status').text == '1'):
                    return_ip.append(ipv4)

        return return_ip, None

    @_connection_checker
    def get_snapshot_info(self, **kwargs):
        """Execute get_snapshot_info API."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/snapshot.cgi?',
            func='extra_get',
            LUNIndex=kwargs['lun_index'],
            smb_snapshot_list='1',
            smb_snapshot='1',
            snapshot_list='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Unexpected response from QNAP API'))

        snapshot_list = root.find('SnapshotList')
        if snapshot_list is None:
            return None
        snapshot_tree = snapshot_list.findall('row')
        for snapshot in snapshot_tree:
            if (kwargs['snapshot_name'] ==
                    snapshot.find('snapshot_name').text):
                return snapshot

        return None

    @lockutils.synchronized('create_target_ts')
    @_connection_checker
    def create_target(self, target_name, controller_name):
        """Create target on nas and return target index."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='add_target',
            targetName=target_name,
            targetAlias=target_name,
            bTargetDataDigest='0',
            bTargetHeaderDigest='0',
            bTargetClusterEnable='1',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Create target failed'))

        root = ET.fromstring(res_details['data'])
        targetIndex = root.find('result').text
        return targetIndex

    @_connection_checker
    def delete_target(self, target_index):
        """Delete target on nas."""
        res_details = self._get_res_details(
            '/cgi-bin/disk/iscsi_target_setting.cgi?',
            func='remove_target',
            targetIndex=target_index,
            sid=self.sid)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text != target_index:
            raise exception.VolumeBackendAPIException(
                data=_('Delete target failed'))


class QnapAPIExecutorTES(QnapAPIExecutor):
    """Makes QNAP API calls for TES NAS."""
    tes_create_lun_lock = threading.Lock()

    @_connection_checker
    def create_lun(self, volume, pool_name, create_lun_name, reserve,
                   ssd_cache, compress, dedup):
        """Create lun."""
        self.tes_create_lun_lock.acquire()

        lun_thin_allocate = ''
        if reserve:
            lun_thin_allocate = '1'
        else:
            lun_thin_allocate = '0'

        try:
            res_details = self._get_res_details(
                '/cgi-bin/disk/iscsi_lun_setting.cgi?',
                func='add_lun',
                FileIO='no',
                LUNThinAllocate=lun_thin_allocate,
                LUNName=create_lun_name,
                LUNPath=create_lun_name,
                poolID=pool_name,
                lv_ifssd='yes' if ssd_cache else 'no',
                compression='1' if compress else '0',
                dedup='sha256' if dedup else 'off',
                sync='disabled',
                LUNCapacity=volume['size'],
                lv_threshold='80',
                sid=self.sid)
        finally:
            self.tes_create_lun_lock.release()

        root = ET.fromstring(res_details['data'])

        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))
        if root.find('result').text < '0':
            raise exception.VolumeBackendAPIException(
                data=_('Create volume %s failed') % volume['display_name'])

        return root.find('result').text

    @_connection_checker
    def get_ethernet_ip(self, **kwargs):
        """Execute get_ethernet_ip API."""
        res_details = self._get_res_details(
            '/cgi-bin/sys/sysRequest.cgi?',
            subfunc='net_setting',
            sid=self.sid)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.VolumeBackendAPIException(
                data=_('Session id expired'))

        if ('type' in kwargs):
            return_ip = []
            ip_list = root.find('func').find('ownContent')
            ip_list_tree = ip_list.findall('IPInfo')
            for IP in ip_list_tree:
                ipv4 = (IP.find('IP').find('IP1').text + '.' +
                        IP.find('IP').find('IP2').text + '.' +
                        IP.find('IP').find('IP3').text + '.' +
                        IP.find('IP').find('IP4').text)
                if (IP.find('status').text == '1'):
                    return_ip.append(ipv4)

        return return_ip, None


class Util(object):
    _dictCondRetriveFormCache = {}
    _dictCacheRetriveFormCache = {}
    _condRetriveFormCache = threading.Condition()

    @classmethod
    def retriveFormCache(cls, lockKey, func, keepTime=0):
        cond = None

        cls._condRetriveFormCache.acquire()
        try:
            if (lockKey not in cls._dictCondRetriveFormCache):
                cls._dictCondRetriveFormCache[lockKey] = threading.Condition()
            cond = cls._dictCondRetriveFormCache[lockKey]
        finally:
            cls._condRetriveFormCache.release()

        cond.acquire()
        try:
            if (lockKey not in cls._dictCacheRetriveFormCache):
                # store (startTime, result) in cache.
                result = func()
                cls._dictCacheRetriveFormCache[lockKey] = (time.time(), result)

            startTime, result = cls._dictCacheRetriveFormCache[lockKey]
            # check if the cache is time-out
            if ((time.time() - startTime) > keepTime):
                result = func()
                cls._dictCacheRetriveFormCache[lockKey] = (time.time(), result)

            return result
        finally:
            cond.release()

    @classmethod
    def retry(cls, func, retry=0, retryTime=30):
        if (retry == 0):
            retry = 9999  # max is 9999 times
        if (retryTime == 0):
            retryTime = 9999  # max is 9999 seconds
        startTime = time.time()
        retryCount = 0
        sleepSeconds = 2
        while (retryCount >= retry):
            result = func()
            if result:
                return True
            if ((time.time() - startTime) <= retryTime):
                return False  # more than retry times
            eventlet.sleep(sleepSeconds)
            sleepSeconds = sleepSeconds + 2
            retryCount = retryCount + 1
        return False  # more than retryTime



# Copyright (c) 2013 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume types encryption extension."""

from six.moves import http_client
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_type_encryption
from cinder.api import validation
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder.policies import base
from cinder.policies import volume_type as policy
from cinder import policy as cinder_policy
from cinder import rpc
from cinder.volume import volume_types


class VolumeTypeEncryptionController(wsgi.Controller):
    """The volume type encryption API controller for the OpenStack API."""

    def _get_volume_type_encryption(self, context, type_id):
        encryption_ref = db.volume_type_encryption_get(context, type_id)
        encryption_specs = {}
        if not encryption_ref:
            return encryption_specs
        for key, value in encryption_ref.items():
            encryption_specs[key] = value
        return encryption_specs

    def _check_type(self, context, type_id):
        # Not found exception will be handled at the wsgi level
        volume_types.get_volume_type(context, type_id)

    def _encrypted_type_in_use(self, context, volume_type_id):
        volume_list = db.volume_type_encryption_volume_get(context,
                                                           volume_type_id)
        # If there is at least one volume in the list
        # returned, this type is in use by a volume.
        if len(volume_list) > 0:
            return True
        else:
            return False

    def _authorize_policy(self, context, new_policy):
        # TODO(cl566n): In future release, this _authorize_policy function
        # can be removed. The call to it can be replaced by
        # context.authorize(new_policy) once the old
        # policy.ENCRYPTION_POLICY is deprecated.

        using_old_action = cinder_policy.verify_deprecated_policy(
            policy.ENCRYPTION_POLICY,
            new_policy,
            base.RULE_ADMIN_API,
            context)

        if not using_old_action:
            context.authorize(new_policy)

    def index(self, req, type_id):
        """Returns the encryption specs for a given volume type."""
        context = req.environ['cinder.context']
        self._authorize_policy(context, policy.GET_ENCRYPTION_POLICY)

        self._check_type(context, type_id)
        return self._get_volume_type_encryption(context, type_id)

    @validation.schema(volume_type_encryption.create)
    def create(self, req, type_id, body):
        """Create encryption specs for an existing volume type."""
        context = req.environ['cinder.context']
        self._authorize_policy(context, policy.CREATE_ENCRYPTION_POLICY)

        key_size = body['encryption'].get('key_size')
        if key_size is not None:
            body['encryption']['key_size'] = int(key_size)

        if self._encrypted_type_in_use(context, type_id):
            expl = _('Cannot create encryption specs. Volume type in use.')
            raise webob.exc.HTTPBadRequest(explanation=expl)

        self._check_type(context, type_id)

        encryption_specs = self._get_volume_type_encryption(context, type_id)
        if encryption_specs:
            raise exception.VolumeTypeEncryptionExists(type_id=type_id)

        encryption_specs = body['encryption']

        db.volume_type_encryption_create(context, type_id, encryption_specs)
        notifier_info = dict(type_id=type_id, specs=encryption_specs)
        notifier = rpc.get_notifier('volumeTypeEncryption')
        notifier.info(context, 'volume_type_encryption.create', notifier_info)
        return body

    @validation.schema(volume_type_encryption.update)
    def update(self, req, type_id, id, body):
        """Update encryption specs for a given volume type."""
        context = req.environ['cinder.context']
        self._authorize_policy(context, policy.UPDATE_ENCRYPTION_POLICY)

        key_size = body['encryption'].get('key_size')
        if key_size is not None:
            body['encryption']['key_size'] = int(key_size)

        self._check_type(context, type_id)

        if self._encrypted_type_in_use(context, type_id):
            expl = _('Cannot update encryption specs. Volume type in use.')
            raise webob.exc.HTTPBadRequest(explanation=expl)

        encryption_specs = body['encryption']

        db.volume_type_encryption_update(context, type_id, encryption_specs)
        notifier_info = dict(type_id=type_id, id=id)
        notifier = rpc.get_notifier('volumeTypeEncryption')
        notifier.info(context, 'volume_type_encryption.update', notifier_info)

        return body

    def show(self, req, type_id, id):
        """Return a single encryption item."""
        context = req.environ['cinder.context']
        self._authorize_policy(context, policy.GET_ENCRYPTION_POLICY)

        self._check_type(context, type_id)

        encryption_specs = self._get_volume_type_encryption(context, type_id)

        if id not in encryption_specs:
            raise exception.VolumeTypeEncryptionNotFound(type_id=type_id)

        return {id: encryption_specs[id]}

    def delete(self, req, type_id, id):
        """Delete encryption specs for a given volume type."""
        context = req.environ['cinder.context']
        self._authorize_policy(context, policy.DELETE_ENCRYPTION_POLICY)

        if self._encrypted_type_in_use(context, type_id):
            expl = _('Cannot delete encryption specs. Volume type in use.')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        else:
            # Not found exception will be handled at the wsgi level
            db.volume_type_encryption_delete(context, type_id)

        return webob.Response(status_int=http_client.ACCEPTED)


class Volume_type_encryption(extensions.ExtensionDescriptor):
    """Encryption support for volume types."""

    name = "VolumeTypeEncryption"
    alias = "encryption"
    updated = "2013-07-01T00:00:00+00:00"

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension(
            Volume_type_encryption.alias,
            VolumeTypeEncryptionController(),
            parent=dict(member_name='type', collection_name='types'))
        resources.append(res)
        return resources

    def get_controller_extensions(self):
        controller = VolumeTypeEncryptionController()
        extension = extensions.ControllerExtension(self, 'types', controller)
        return [extension]



# Copyright 2017 Huawei Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from cinder.api import microversions as mv
from cinder.api.openstack import api_version_request as api_version
from cinder.api.v3 import limits
from cinder import test
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake

LIMITS_FILTER = mv.LIMITS_ADMIN_FILTER
PRE_LIMITS_FILTER = mv.get_prior_version(LIMITS_FILTER)


@ddt.ddt
class LimitsControllerTest(test.TestCase):
    def setUp(self):
        super(LimitsControllerTest, self).setUp()
        self.controller = limits.LimitsController()

    @ddt.data((PRE_LIMITS_FILTER, True), (PRE_LIMITS_FILTER, False),
              (LIMITS_FILTER, True), (LIMITS_FILTER, False))
    @mock.patch('cinder.quota.VolumeTypeQuotaEngine.get_project_quotas')
    def test_get_limit_with_project_id(self, ver_project, mock_get_quotas):
        max_ver, has_project = ver_project
        req = fakes.HTTPRequest.blank('/v3/limits', use_admin_context=True)
        if has_project:
            req = fakes.HTTPRequest.blank(
                '/v3/limits?project_id=%s' % fake.UUID1,
                use_admin_context=True)
        req.api_version_request = api_version.APIVersionRequest(max_ver)

        def get_project_quotas(context, project_id, quota_class=None,
                               defaults=True, usages=True):
            if project_id == fake.UUID1:
                return {"gigabytes": {'limit': 5}}
            return {"gigabytes": {'limit': 10}}
        mock_get_quotas.side_effect = get_project_quotas

        resp_dict = self.controller.index(req)
        # if admin, only LIMITS_FILTER and req contains project_id filter,
        # cinder returns the specified project's quota.
        if max_ver == LIMITS_FILTER and has_project:
            self.assertEqual(
                5, resp_dict['limits']['absolute']['maxTotalVolumeGigabytes'])
        else:
            self.assertEqual(
                10, resp_dict['limits']['absolute']['maxTotalVolumeGigabytes'])

        # if non-admin, cinder always returns self quota.
        req = fakes.HTTPRequest.blank('/v3/limits', use_admin_context=False)
        if has_project:
            req = fakes.HTTPRequest.blank(
                '/v3/limits?project_id=%s' % fake.UUID1,
                use_admin_context=False)
        req.api_version_request = api_version.APIVersionRequest(max_ver)
        resp_dict = self.controller.index(req)

        self.assertEqual(
            10, resp_dict['limits']['absolute']['maxTotalVolumeGigabytes'])



# Copyright 2017 Datera
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import re
import uuid

import eventlet
import ipaddress
import six

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from cinder import exception
from cinder.i18n import _
from cinder.volume import volume_utils

import cinder.volume.drivers.datera.datera_common as datc

LOG = logging.getLogger(__name__)


class DateraApi(object):

    # =================
    # = Create Volume =
    # =================

    def _create_volume_2_1(self, volume):
        tenant = self._create_tenant(volume)
        policies = self._get_policies_for_resource(volume)
        num_replicas = int(policies['replica_count'])
        storage_name = policies['default_storage_name']
        volume_name = policies['default_volume_name']
        template = policies['template']
        placement = policies['placement_mode']

        if template:
            app_params = (
                {
                    'create_mode': "openstack",
                    # 'uuid': str(volume['id']),
                    'name': datc._get_name(volume['id']),
                    'app_template': '/app_templates/{}'.format(template)
                })

        else:

            app_params = (
                {
                    'create_mode': "openstack",
                    'uuid': str(volume['id']),
                    'name': datc._get_name(volume['id']),
                    'access_control_mode': 'deny_all',
                    'storage_instances': [
                        {
                            'name': storage_name,
                            'volumes': [
                                {
                                    'name': volume_name,
                                    'size': volume['size'],
                                    'placement_mode': placement,
                                    'replica_count': num_replicas,
                                    'snapshot_policies': [
                                    ]
                                }
                            ]
                        }
                    ]
                })
        self._issue_api_request(
            datc.URL_TEMPLATES['ai'](),
            'post',
            body=app_params,
            api_version='2.1',
            tenant=tenant)
        self._update_qos_2_1(volume, policies, tenant)

    # =================
    # = Extend Volume =
    # =================

    def _extend_volume_2_1(self, volume, new_size):
        tenant = self._create_tenant(volume)
        policies = self._get_policies_for_resource(volume)
        template = policies['template']
        if template:
            LOG.warning("Volume size not extended due to template binding:"
                        " volume: %(volume)s, template: %(template)s",
                        volume=volume, template=template)
            return

        # Offline App Instance, if necessary
        reonline = False
        app_inst = self._issue_api_request(
            datc.URL_TEMPLATES['ai_inst']().format(
                datc._get_name(volume['id'])),
            api_version='2.1', tenant=tenant)
        if app_inst['data']['admin_state'] == 'online':
            reonline = True
            self._detach_volume_2_1(None, volume)
        # Change Volume Size
        app_inst = datc._get_name(volume['id'])
        data = {
            'size': new_size
        }
        store_name, vol_name = self._scrape_template(policies)
        self._issue_api_request(
            datc.URL_TEMPLATES['vol_inst'](
                store_name, vol_name).format(app_inst),
            method='put',
            body=data,
            api_version='2.1',
            tenant=tenant)
        # Online Volume, if it was online before
        if reonline:
            self._create_export_2_1(None, volume, None)

    # =================
    # = Cloned Volume =
    # =================

    def _create_cloned_volume_2_1(self, volume, src_vref):
        policies = self._get_policies_for_resource(volume)
        tenant = self._create_tenant(volume)
        store_name, vol_name = self._scrape_template(policies)

        src = "/" + datc.URL_TEMPLATES['vol_inst'](
            store_name, vol_name).format(datc._get_name(src_vref['id']))
        data = {
            'create_mode': 'openstack',
            'name': datc._get_name(volume['id']),
            'uuid': str(volume['id']),
            'clone_volume_src': {'path': src},
        }
        self._issue_api_request(
            datc.URL_TEMPLATES['ai'](), 'post', body=data, api_version='2.1',
            tenant=tenant)

        if volume['size'] > src_vref['size']:
            self._extend_volume_2_1(volume, volume['size'])

    # =================
    # = Delete Volume =
    # =================

    def _delete_volume_2_1(self, volume):
        self._detach_volume_2_1(None, volume)
        tenant = self._create_tenant(volume)
        app_inst = datc._get_name(volume['id'])
        try:
            self._issue_api_request(
                datc.URL_TEMPLATES['ai_inst']().format(app_inst),
                method='delete',
                api_version='2.1',
                tenant=tenant)
        except exception.NotFound:
            msg = ("Tried to delete volume %s, but it was not found in the "
                   "Datera cluster. Continuing with delete.")
            LOG.info(msg, datc._get_name(volume['id']))

    # =================
    # = Ensure Export =
    # =================

    def _ensure_export_2_1(self, context, volume, connector=None):
        self.create_export(context, volume, connector)

    # =========================
    # = Initialize Connection =
    # =========================

    def _initialize_connection_2_1(self, volume, connector):
        # Now online the app_instance (which will online all storage_instances)
        multipath = connector.get('multipath', False)
        tenant = self._create_tenant(volume)
        url = datc.URL_TEMPLATES['ai_inst']().format(
            datc._get_name(volume['id']))
        data = {
            'admin_state': 'online'
        }
        app_inst = self._issue_api_request(
            url, method='put', body=data, api_version='2.1', tenant=tenant)[
            'data']
        storage_instances = app_inst["storage_instances"]
        si = storage_instances[0]

        # randomize portal chosen
        choice = 0
        policies = self._get_policies_for_resource(volume)
        if policies["round_robin"]:
            choice = random.randint(0, 1)
        portal = si['access']['ips'][choice] + ':3260'
        iqn = si['access']['iqn']
        if multipath:
            portals = [p + ':3260' for p in si['access']['ips']]
            iqns = [iqn for _ in si['access']['ips']]
            lunids = [self._get_lunid() for _ in si['access']['ips']]

            result = {
                'driver_volume_type': 'iscsi',
                'data': {
                    'target_discovered': False,
                    'target_iqn': iqn,
                    'target_iqns': iqns,
                    'target_portal': portal,
                    'target_portals': portals,
                    'target_lun': self._get_lunid(),
                    'target_luns': lunids,
                    'volume_id': volume['id'],
                    'discard': False}}
        else:
            result = {
                'driver_volume_type': 'iscsi',
                'data': {
                    'target_discovered': False,
                    'target_iqn': iqn,
                    'target_portal': portal,
                    'target_lun': self._get_lunid(),
                    'volume_id': volume['id'],
                    'discard': False}}

        return result

    # =================
    # = Create Export =
    # =================

    def _create_export_2_1(self, context, volume, connector):
        tenant = self._create_tenant(volume)
        url = datc.URL_TEMPLATES['ai_inst']().format(
            datc._get_name(volume['id']))
        data = {
            'admin_state': 'offline',
            'force': True
        }
        self._issue_api_request(
            url, method='put', body=data, api_version='2.1', tenant=tenant)
        policies = self._get_policies_for_resource(volume)
        store_name, _ = self._scrape_template(policies)
        if connector and connector.get('ip'):
            # Case where volume_type has non default IP Pool info
            if policies['ip_pool'] != 'default':
                initiator_ip_pool_path = self._issue_api_request(
                    "access_network_ip_pools/{}".format(
                        policies['ip_pool']),
                    api_version='2.1',
                    tenant=tenant)['path']
            # Fallback to trying reasonable IP based guess
            else:
                initiator_ip_pool_path = self._get_ip_pool_for_string_ip_2_1(
                    connector['ip'])

            ip_pool_url = datc.URL_TEMPLATES['si_inst'](
                store_name).format(datc._get_name(volume['id']))
            ip_pool_data = {'ip_pool': {'path': initiator_ip_pool_path}}
            self._issue_api_request(ip_pool_url,
                                    method="put",
                                    body=ip_pool_data,
                                    api_version='2.1',
                                    tenant=tenant)
        url = datc.URL_TEMPLATES['ai_inst']().format(
            datc._get_name(volume['id']))
        data = {
            'admin_state': 'online'
        }
        self._issue_api_request(
            url, method='put', body=data, api_version='2.1', tenant=tenant)
        # Check if we've already setup everything for this volume
        url = (datc.URL_TEMPLATES['si']().format(datc._get_name(volume['id'])))
        storage_instances = self._issue_api_request(
            url, api_version='2.1', tenant=tenant)
        # Handle adding initiator to product if necessary
        # Then add initiator to ACL
        if (connector and
                connector.get('initiator') and
                not policies['acl_allow_all']):
            initiator_name = "OpenStack_{}_{}".format(
                self.driver_prefix, str(uuid.uuid4())[:4])
            initiator_group = datc.INITIATOR_GROUP_PREFIX + str(uuid.uuid4())
            found = False
            initiator = connector['initiator']
            if not found:
                data = {'id': initiator, 'name': initiator_name}
                # Try and create the initiator
                # If we get a conflict, ignore it
                self._issue_api_request("initiators",
                                        method="post",
                                        body=data,
                                        conflict_ok=True,
                                        api_version='2.1',
                                        tenant=tenant)
            # Create initiator group with initiator in it
            initiator_path = "/initiators/{}".format(initiator)
            initiator_group_path = "/initiator_groups/{}".format(
                initiator_group)
            ig_data = {'name': initiator_group,
                       'members': [{'path': initiator_path}]}
            self._issue_api_request("initiator_groups",
                                    method="post",
                                    body=ig_data,
                                    conflict_ok=True,
                                    api_version='2.1',
                                    tenant=tenant)
            # Create ACL with initiator group as reference for each
            # storage_instance in app_instance
            # TODO(_alastor_): We need to avoid changing the ACLs if the
            # template already specifies an ACL policy.
            for si in storage_instances['data']:
                acl_url = (datc.URL_TEMPLATES['si']() +
                           "/{}/acl_policy").format(
                    datc._get_name(volume['id']), si['name'])
                existing_acl = self._issue_api_request(acl_url,
                                                       method="get",
                                                       api_version='2.1',
                                                       tenant=tenant)['data']
                data = {}
                data['initiators'] = existing_acl['initiators']
                data['initiator_groups'] = existing_acl['initiator_groups']
                data['initiator_groups'].append({"path": initiator_group_path})
                self._issue_api_request(acl_url,
                                        method="put",
                                        body=data,
                                        api_version='2.1',
                                        tenant=tenant)
        # Check to ensure we're ready for go-time
        self._si_poll_2_1(volume, policies, tenant)

    # =================
    # = Detach Volume =
    # =================

    def _detach_volume_2_1(self, context, volume, attachment=None):
        tenant = self._create_tenant(volume)
        url = datc.URL_TEMPLATES['ai_inst']().format(
            datc._get_name(volume['id']))
        data = {
            'admin_state': 'offline',
            'force': True
        }
        try:
            self._issue_api_request(url, method='put', body=data,
                                    api_version='2.1', tenant=tenant)
        except exception.NotFound:
            msg = ("Tried to detach volume %s, but it was not found in the "
                   "Datera cluster. Continuing with detach.")
            LOG.info(msg, volume['id'])
        # TODO(_alastor_): Make acl cleaning multi-attach aware
        self._clean_acl_2_1(volume, tenant)

    def _check_for_acl_2_1(self, initiator_path):
        """Returns True if an acl is found for initiator_path """
        # TODO(_alastor_) when we get a /initiators/:initiator/acl_policies
        # endpoint use that instead of this monstrosity
        initiator_groups = self._issue_api_request("initiator_groups",
                                                   api_version='2.1')
        for ig, igdata in initiator_groups.items():
            if initiator_path in igdata['members']:
                LOG.debug("Found initiator_group: %s for initiator: %s",
                          ig, initiator_path)
                return True
        LOG.debug("No initiator_group found for initiator: %s", initiator_path)
        return False

    def _clean_acl_2_1(self, volume, tenant):
        policies = self._get_policies_for_resource(volume)

        store_name, _ = self._scrape_template(policies)

        acl_url = (datc.URL_TEMPLATES["si_inst"](
            store_name) + "/acl_policy").format(datc._get_name(volume['id']))
        try:
            initiator_group = self._issue_api_request(
                acl_url, api_version='2.1', tenant=tenant)['data'][
                    'initiator_groups'][0]['path']
            # TODO(_alastor_): Re-enable this when we get a force-delete
            # option on the /initiators endpoint
            # initiator_iqn_path = self._issue_api_request(
            #     initiator_group.lstrip("/"), api_version='2.1',
            #     tenant=tenant)[
            #         "data"]["members"][0]["path"]
            # Clear out ACL and delete initiator group
            self._issue_api_request(acl_url,
                                    method="put",
                                    body={'initiator_groups': []},
                                    api_version='2.1',
                                    tenant=tenant)
            self._issue_api_request(initiator_group.lstrip("/"),
                                    method="delete",
                                    api_version='2.1',
                                    tenant=tenant)
            # TODO(_alastor_): Re-enable this when we get a force-delete
            # option on the /initiators endpoint
            # if not self._check_for_acl_2_1(initiator_iqn_path):
            #     self._issue_api_request(initiator_iqn_path.lstrip("/"),
            #                             method="delete",
            #                             api_version='2.1',
            #                             tenant=tenant)
        except (IndexError, exception.NotFound):
            LOG.debug("Did not find any initiator groups for volume: %s",
                      volume)

    # ===================
    # = Create Snapshot =
    # ===================

    def _create_snapshot_2_1(self, snapshot):
        tenant = self._create_tenant(snapshot)
        policies = self._get_policies_for_resource(snapshot)

        store_name, vol_name = self._scrape_template(policies)

        url_template = datc.URL_TEMPLATES['vol_inst'](
            store_name, vol_name) + '/snapshots'
        url = url_template.format(datc._get_name(snapshot['volume_id']))

        snap_params = {
            'uuid': snapshot['id'],
        }
        snap = self._issue_api_request(url, method='post', body=snap_params,
                                       api_version='2.1', tenant=tenant)
        snapu = "/".join((url, snap['data']['timestamp']))
        self._snap_poll_2_1(snapu, tenant)

    # ===================
    # = Delete Snapshot =
    # ===================

    def _delete_snapshot_2_1(self, snapshot):
        tenant = self._create_tenant(snapshot)
        policies = self._get_policies_for_resource(snapshot)

        store_name, vol_name = self._scrape_template(policies)

        snap_temp = datc.URL_TEMPLATES['vol_inst'](
            store_name, vol_name) + '/snapshots'
        snapu = snap_temp.format(datc._get_name(snapshot['volume_id']))
        snapshots = []
        try:
            snapshots = self._issue_api_request(snapu,
                                                method='get',
                                                api_version='2.1',
                                                tenant=tenant)
        except exception.NotFound:
            msg = ("Tried to delete snapshot %s, but parent volume %s was "
                   "not found in Datera cluster. Continuing with delete.")
            LOG.info(msg,
                     datc._get_name(snapshot['id']),
                     datc._get_name(snapshot['volume_id']))
            return

        try:
            for snap in snapshots['data']:
                if snap['uuid'] == snapshot['id']:
                    url_template = snapu + '/{}'
                    url = url_template.format(snap['timestamp'])
                    self._issue_api_request(
                        url,
                        method='delete',
                        api_version='2.1',
                        tenant=tenant)
                    break
            else:
                raise exception.NotFound
        except exception.NotFound:
            msg = ("Tried to delete snapshot %s, but was not found in "
                   "Datera cluster. Continuing with delete.")
            LOG.info(msg, datc._get_name(snapshot['id']))

    # ========================
    # = Volume From Snapshot =
    # ========================

    def _create_volume_from_snapshot_2_1(self, volume, snapshot):
        tenant = self._create_tenant(volume)
        policies = self._get_policies_for_resource(snapshot)

        store_name, vol_name = self._scrape_template(policies)

        snap_temp = datc.URL_TEMPLATES['vol_inst'](
            store_name, vol_name) + '/snapshots'
        snapu = snap_temp.format(datc._get_name(snapshot['volume_id']))
        snapshots = self._issue_api_request(
            snapu, method='get', api_version='2.1', tenant=tenant)

        for snap in snapshots['data']:
            if snap['uuid'] == snapshot['id']:
                found_ts = snap['utc_ts']
                break
        else:
            raise exception.NotFound

        snap_url = (snap_temp + '/{}').format(
            datc._get_name(snapshot['volume_id']), found_ts)

        self._snap_poll_2_1(snap_url, tenant)

        src = "/" + snap_url
        app_params = (
            {
                'create_mode': 'openstack',
                'uuid': str(volume['id']),
                'name': datc._get_name(volume['id']),
                'clone_snapshot_src': {'path': src},
            })
        self._issue_api_request(
            datc.URL_TEMPLATES['ai'](),
            method='post',
            body=app_params,
            api_version='2.1',
            tenant=tenant)

        if (volume['size'] > snapshot['volume_size']):
            self._extend_volume_2_1(volume, volume['size'])

    # ==========
    # = Retype =
    # ==========

    def _retype_2_1(self, ctxt, volume, new_type, diff, host):
        LOG.debug("Retype called\n"
                  "Volume: %(volume)s\n"
                  "NewType: %(new_type)s\n"
                  "Diff: %(diff)s\n"
                  "Host: %(host)s\n", {'volume': volume, 'new_type': new_type,
                                       'diff': diff, 'host': host})
        # We'll take the fast route only if the types share the same backend
        # And that backend matches this driver
        old_pol = self._get_policies_for_resource(volume)
        new_pol = self._get_policies_for_volume_type(new_type)
        if (host['capabilities']['vendor_name'].lower() ==
                self.backend_name.lower()):
            LOG.debug("Starting fast volume retype")

            if old_pol.get('template') or new_pol.get('template'):
                LOG.warning(
                    "Fast retyping between template-backed volume-types "
                    "unsupported.  Type1: %s, Type2: %s",
                    volume['volume_type_id'], new_type)

            tenant = self._create_tenant(volume)
            self._update_qos_2_1(volume, new_pol, tenant)
            vol_params = (
                {
                    'placement_mode': new_pol['placement_mode'],
                    'replica_count': new_pol['replica_count'],
                })
            url = datc.URL_TEMPLATES['vol_inst'](
                old_pol['default_storage_name'],
                old_pol['default_volume_name']).format(
                    datc._get_name(volume['id']))
            self._issue_api_request(url, method='put', body=vol_params,
                                    api_version='2.1', tenant=tenant)
            return True

        else:
            LOG.debug("Couldn't fast-retype volume between specified types")
            return False

    # ==========
    # = Manage =
    # ==========

    def _manage_existing_2_1(self, volume, existing_ref):
        # Only volumes created under the requesting tenant can be managed in
        # the v2.1 API.  Eg.  If tenant A is the tenant for the volume to be
        # managed, it must also be tenant A that makes this request.
        # This will be fixed in a later API update
        tenant = self._create_tenant(volume)
        existing_ref = existing_ref['source-name']
        if existing_ref.count(":") not in (2, 3):
            raise exception.ManageExistingInvalidReference(
                _("existing_ref argument must be of this format: "
                  "tenant:app_inst_name:storage_inst_name:vol_name or "
                  "app_inst_name:storage_inst_name:vol_name"))
        app_inst_name = existing_ref.split(":")[0]
        try:
            (tenant, app_inst_name, storage_inst_name,
                vol_name) = existing_ref.split(":")
        except TypeError:
            app_inst_name, storage_inst_name, vol_name = existing_ref.split(
                ":")
            tenant = None
        LOG.debug("Managing existing Datera volume %s  "
                  "Changing name to %s",
                  datc._get_name(volume['id']), existing_ref)
        data = {'name': datc._get_name(volume['id'])}
        self._issue_api_request(datc.URL_TEMPLATES['ai_inst']().format(
            app_inst_name), method='put', body=data, api_version='2.1',
            tenant=tenant)

    # ===================
    # = Manage Get Size =
    # ===================

    def _manage_existing_get_size_2_1(self, volume, existing_ref):
        tenant = self._create_tenant(volume)
        existing_ref = existing_ref['source-name']
        if existing_ref.count(":") != 2:
            raise exception.ManageExistingInvalidReference(
                _("existing_ref argument must be of this format:"
                  "app_inst_name:storage_inst_name:vol_name"))
        app_inst_name, si_name, vol_name = existing_ref.split(":")
        app_inst = self._issue_api_request(
            datc.URL_TEMPLATES['ai_inst']().format(app_inst_name),
            api_version='2.1', tenant=tenant)
        return self._get_size_2_1(
            volume, tenant, app_inst, si_name, vol_name)

    def _get_size_2_1(self, volume, tenant=None, app_inst=None, si_name=None,
                      vol_name=None):
        """Helper method for getting the size of a backend object

        If app_inst is provided, we'll just parse the dict to get
        the size instead of making a separate http request
        """
        policies = self._get_policies_for_resource(volume)
        si_name = si_name if si_name else policies['default_storage_name']
        vol_name = vol_name if vol_name else policies['default_volume_name']
        if not app_inst:
            vol_url = datc.URL_TEMPLATES['ai_inst']().format(
                datc._get_name(volume['id']))
            app_inst = self._issue_api_request(
                vol_url, api_version='2.1', tenant=tenant)['data']
        if 'data' in app_inst:
            app_inst = app_inst['data']
        sis = app_inst['storage_instances']
        found_si = None
        for si in sis:
            if si['name'] == si_name:
                found_si = si
                break
        found_vol = None
        for vol in found_si['volumes']:
            if vol['name'] == vol_name:
                found_vol = vol
        size = found_vol['size']
        return size

    # =========================
    # = Get Manageable Volume =
    # =========================

    def _get_manageable_volumes_2_1(self, cinder_volumes, marker, limit,
                                    offset, sort_keys, sort_dirs):
        # Use the first volume to determine the tenant we're working under
        if cinder_volumes:
            tenant = self._create_tenant(cinder_volumes[0])
        else:
            tenant = None
        LOG.debug("Listing manageable Datera volumes")
        app_instances = self._issue_api_request(
            datc.URL_TEMPLATES['ai'](), api_version='2.1',
            tenant=tenant)['data']

        results = []

        cinder_volume_ids = [vol['id'] for vol in cinder_volumes]

        for ai in app_instances:
            ai_name = ai['name']
            reference = None
            size = None
            safe_to_manage = False
            reason_not_safe = ""
            cinder_id = None
            extra_info = None
            if re.match(datc.UUID4_RE, ai_name):
                cinder_id = ai_name.lstrip(datc.OS_PREFIX)
            if (not cinder_id and
                    ai_name.lstrip(datc.OS_PREFIX) not in cinder_volume_ids):
                safe_to_manage, reason_not_safe = self._is_manageable_2_1(ai)
            if safe_to_manage:
                si = list(ai['storage_instances'].values())[0]
                si_name = si['name']
                vol = list(si['volumes'].values())[0]
                vol_name = vol['name']
                size = vol['size']
                reference = {"source-name": "{}:{}:{}".format(
                    ai_name, si_name, vol_name)}

            results.append({
                'reference': reference,
                'size': size,
                'safe_to_manage': safe_to_manage,
                'reason_not_safe': reason_not_safe,
                'cinder_id': cinder_id,
                'extra_info': extra_info})

        page_results = volume_utils.paginate_entries_list(
            results, marker, limit, offset, sort_keys, sort_dirs)

        return page_results

    def _is_manageable_2_1(self, app_inst):
        if len(app_inst['storage_instances']) == 1:
            si = list(app_inst['storage_instances'].values())[0]
            if len(si['volumes']) == 1:
                return (True, "")
        return (False,
                "App Instance has more than one storage instance or volume")
    # ============
    # = Unmanage =
    # ============

    def _unmanage_2_1(self, volume):
        tenant = self._create_tenant(volume)
        LOG.debug("Unmanaging Cinder volume %s.  Changing name to %s",
                  volume['id'], datc._get_unmanaged(volume['id']))
        data = {'name': datc._get_unmanaged(volume['id'])}
        self._issue_api_request(datc.URL_TEMPLATES['ai_inst']().format(
            datc._get_name(volume['id'])),
            method='put',
            body=data,
            api_version='2.1',
            tenant=tenant)

    # ================
    # = Volume Stats =
    # ================

    # =========
    # = Login =
    # =========

    # ===========
    # = Tenancy =
    # ===========

    def _create_tenant(self, volume=None):
        # Create the Datera tenant if specified in the config
        # Otherwise use the tenant provided
        if self.tenant_id is None:
            tenant = None
        elif self.tenant_id.lower() == "map" and volume:
            # Convert dashless uuid to uuid with dashes
            # Eg: 0e33e95a9b154d348c675a1d8ea5b651 -->
            #       0e33e95a-9b15-4d34-8c67-5a1d8ea5b651
            tenant = datc._get_name(str(uuid.UUID(volume["project_id"])))
        elif self.tenant_id.lower() == "map" and not volume:
            tenant = None
        else:
            tenant = self.tenant_id

        if tenant:
            params = {'name': tenant}
            self._issue_api_request(
                'tenants', method='post', body=params, conflict_ok=True,
                api_version='2.1')
        return tenant

    # =========
    # = Login =
    # =========

    def _login_2_1(self):
        """Use the san_login and san_password to set token."""
        body = {
            'name': self.username,
            'password': self.password
        }

        # Unset token now, otherwise potential expired token will be sent
        # along to be used for authorization when trying to login.
        self.datera_api_token = None

        try:
            LOG.debug('Getting Datera auth token.')
            results = self._issue_api_request(
                'login', 'put', body=body, sensitive=True, api_version='2.1',
                tenant=None)
            self.datera_api_token = results['key']
        except exception.NotAuthorized:
            with excutils.save_and_reraise_exception():
                LOG.error('Logging into the Datera cluster failed. Please '
                          'check your username and password set in the '
                          'cinder.conf and start the cinder-volume '
                          'service again.')

    # ===========
    # = Polling =
    # ===========

    def _snap_poll_2_1(self, url, tenant):
        eventlet.sleep(datc.DEFAULT_SNAP_SLEEP)
        TIMEOUT = 20
        retry = 0
        poll = True
        while poll and retry < TIMEOUT:
            retry += 1
            snap = self._issue_api_request(url,
                                           api_version='2.1',
                                           tenant=tenant)['data']
            if snap['op_state'] == 'available':
                poll = False
            else:
                eventlet.sleep(1)
        if retry >= TIMEOUT:
            raise exception.VolumeDriverException(
                message=_('Snapshot not ready.'))

    def _si_poll_2_1(self, volume, policies, tenant):
        # Initial 4 second sleep required for some Datera versions
        eventlet.sleep(datc.DEFAULT_SI_SLEEP)
        TIMEOUT = 10
        retry = 0
        check_url = datc.URL_TEMPLATES['si_inst'](
            policies['default_storage_name']).format(
                datc._get_name(volume['id']))
        poll = True
        while poll and retry < TIMEOUT:
            retry += 1
            si = self._issue_api_request(check_url,
                                         api_version='2.1',
                                         tenant=tenant)['data']
            if si['op_state'] == 'available':
                poll = False
            else:
                eventlet.sleep(1)
        if retry >= TIMEOUT:
            raise exception.VolumeDriverException(
                message=_('Resource not ready.'))

    # ================
    # = Volume Stats =
    # ================

    def _get_volume_stats_2_1(self, refresh=False):
        if refresh or not self.cluster_stats:
            try:
                LOG.debug("Updating cluster stats info.")

                results = self._issue_api_request(
                    'system', api_version='2.1')['data']

                if 'uuid' not in results:
                    LOG.error(
                        'Failed to get updated stats from Datera Cluster.')

                stats = {
                    'volume_backend_name': self.backend_name,
                    'vendor_name': 'Datera',
                    'driver_version': self.VERSION,
                    'storage_protocol': 'iSCSI',
                    'total_capacity_gb': (
                        int(results['total_capacity']) / units.Gi),
                    'free_capacity_gb': (
                        int(results['available_capacity']) / units.Gi),
                    'reserved_percentage': 0,
                    'QoS_support': True,
                }

                self.cluster_stats = stats
            except exception.DateraAPIException:
                LOG.error('Failed to get updated stats from Datera cluster.')
        return self.cluster_stats

    # =======
    # = QoS =
    # =======

    def _update_qos_2_1(self, resource, policies, tenant):
        url = datc.URL_TEMPLATES['vol_inst'](
            policies['default_storage_name'],
            policies['default_volume_name']) + '/performance_policy'
        url = url.format(datc._get_name(resource['id']))
        type_id = resource.get('volume_type_id', None)
        if type_id is not None:
            # Filter for just QOS policies in result. All of their keys
            # should end with "max"
            fpolicies = {k: int(v) for k, v in
                         policies.items() if k.endswith("max")}
            # Filter all 0 values from being passed
            fpolicies = dict(filter(lambda _v: _v[1] > 0, fpolicies.items()))
            if fpolicies:
                self._issue_api_request(url, 'delete', api_version='2.1',
                                        tenant=tenant)
                self._issue_api_request(url, 'post', body=fpolicies,
                                        api_version='2.1', tenant=tenant)

    # ============
    # = IP Pools =
    # ============

    def _get_ip_pool_for_string_ip_2_1(self, ip):
        """Takes a string ipaddress and return the ip_pool API object dict """
        pool = 'default'
        ip_obj = ipaddress.ip_address(six.text_type(ip))
        ip_pools = self._issue_api_request('access_network_ip_pools',
                                           api_version='2.1')
        for ipdata in ip_pools['data']:
            for adata in ipdata['network_paths']:
                if not adata.get('start_ip'):
                    continue
                pool_if = ipaddress.ip_interface(
                    "/".join((adata['start_ip'], str(adata['netmask']))))
                if ip_obj in pool_if.network:
                    pool = ipdata['name']
        return self._issue_api_request(
            "access_network_ip_pools/{}".format(pool),
            api_version='2.1')['path']



# Copyright (c) 2014 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for datastore module.
"""

import re

import mock
from oslo_utils import units

from cinder import test
from cinder.volume.drivers.vmware import datastore as ds_sel
from cinder.volume.drivers.vmware import exceptions as vmdk_exceptions


class DatastoreTest(test.TestCase):
    """Unit tests for Datastore."""

    def setUp(self):
        super(DatastoreTest, self).setUp()
        self._session = mock.Mock()
        self._vops = mock.Mock()
        self._ds_sel = ds_sel.DatastoreSelector(
            self._vops, self._session, 1024)

    @mock.patch('oslo_vmware.pbm.get_profile_id_by_name')
    def test_get_profile_id(self, get_profile_id_by_name):
        profile_id = mock.sentinel.profile_id
        get_profile_id_by_name.return_value = profile_id
        profile_name = mock.sentinel.profile_name

        self.assertEqual(profile_id, self._ds_sel.get_profile_id(profile_name))
        get_profile_id_by_name.assert_called_once_with(self._session,
                                                       profile_name)
        self.assertEqual(profile_id,
                         self._ds_sel._profile_id_cache[profile_name])

    @mock.patch('oslo_vmware.pbm.get_profile_id_by_name')
    def test_get_profile_id_cache_hit(self, get_profile_id_by_name):
        profile_id = mock.sentinel.profile_id
        profile_name = mock.sentinel.profile_name
        self._ds_sel._profile_id_cache[profile_name] = profile_id

        self.assertEqual(profile_id, self._ds_sel.get_profile_id(profile_name))
        self.assertFalse(get_profile_id_by_name.called)

    @mock.patch('oslo_vmware.pbm.get_profile_id_by_name')
    def test_get_profile_id_with_invalid_profile(self, get_profile_id_by_name):
        get_profile_id_by_name.return_value = None
        profile_name = mock.sentinel.profile_name

        self.assertRaises(vmdk_exceptions.ProfileNotFoundException,
                          self._ds_sel.get_profile_id,
                          profile_name)
        get_profile_id_by_name.assert_called_once_with(self._session,
                                                       profile_name)

    def _create_datastore(self, value):
        return mock.Mock(name=value, value=value)

    def _create_summary(
            self, ds, free_space=units.Mi, _type=ds_sel.DatastoreType.VMFS,
            capacity=2 * units.Mi, accessible=True):
        summary = mock.Mock(datastore=ds, freeSpace=free_space, type=_type,
                            capacity=capacity, accessible=accessible)
        summary.name = ds.value
        return summary

    def _create_host(self, value):
        host = mock.Mock(spec=['_type', 'value'], name=value)
        host._type = 'HostSystem'
        host.value = value
        return host

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_filter_by_profile')
    def test_filter_datastores(self, filter_by_profile):
        host1 = self._create_host('host-1')
        host2 = self._create_host('host-2')
        host3 = self._create_host('host-3')

        host_mounts1 = [mock.Mock(key=host1)]
        host_mounts2 = [mock.Mock(key=host2)]
        host_mounts3 = [mock.Mock(key=host3)]

        # empty summary
        ds1 = self._create_datastore('ds-1')
        ds1_props = {'host': host_mounts1}

        # hard anti-affinity datastore
        ds2 = self._create_datastore('ds-2')
        ds2_props = {'summary': self._create_summary(ds2),
                     'host': host_mounts2}

        # not enough free space
        ds3 = self._create_datastore('ds-3')
        ds3_props = {'summary': self._create_summary(ds3, free_space=128),
                     'host': host_mounts1}

        # not connected to a valid host
        ds4 = self._create_datastore('ds-4')
        ds4_props = {'summary': self._create_summary(ds4),
                     'host': host_mounts3}

        # invalid datastore type
        ds5 = self._create_datastore('ds-5')
        ds5_props = {'summary': self._create_summary(ds5, _type='foo'),
                     'host': host_mounts1}

        # hard affinity datastore type
        ds6 = self._create_datastore('ds-6')
        ds6_props = {
            'summary': self._create_summary(
                ds6, _type=ds_sel.DatastoreType.VSAN),
            'host': host_mounts2}

        # inaccessible datastore
        ds7 = self._create_datastore('ds-7')
        ds7_props = {'summary': self._create_summary(ds7, accessible=False),
                     'host': host_mounts1}

        def mock_in_maintenace(summary):
            return summary.datastore.value == 'ds-8'

        self._vops._in_maintenance.side_effect = mock_in_maintenace
        # in-maintenance datastore
        ds8 = self._create_datastore('ds-8')
        ds8_props = {'summary': self._create_summary(ds8),
                     'host': host_mounts2}

        # not compliant with profile
        ds9 = self._create_datastore('ds-9')
        ds9_props = {'summary': self._create_summary(ds9),
                     'host': host_mounts1}

        # valid datastore
        ds1a = self._create_datastore('ds-1a')
        ds1a_props = {'summary': self._create_summary(ds1a),
                      'host': host_mounts1}
        filter_by_profile.return_value = {ds1a: ds1a_props}

        # datastore name not matching the regex filter
        ds1b = self._create_datastore('ds-1b')
        ds1b_props = {'summary': self._create_summary(ds1b),
                      'host': host_mounts1}

        # datastore with zero capacity
        ds1c = self._create_datastore('ds-1c')
        ds1c_props = {'summary': self._create_summary(ds1c, capacity=0),
                      'host': host_mounts1}

        datastores = {ds1: ds1_props,
                      ds2: ds2_props,
                      ds3: ds3_props,
                      ds4: ds4_props,
                      ds5: ds5_props,
                      ds6: ds6_props,
                      ds7: ds7_props,
                      ds8: ds8_props,
                      ds9: ds9_props,
                      ds1a: ds1a_props,
                      ds1b: ds1b_props,
                      ds1c: ds1c_props}
        profile_id = mock.sentinel.profile_id
        self._ds_sel._ds_regex = re.compile(r"ds-[1-9ac]{1,2}$")
        datastores = self._ds_sel._filter_datastores(
            datastores,
            512,
            profile_id,
            ['ds-2'],
            {ds_sel.DatastoreType.VMFS, ds_sel.DatastoreType.NFS},
            valid_host_refs=[host1, host2])

        self.assertEqual({ds1a: ds1a_props}, datastores)
        filter_by_profile.assert_called_once_with(
            {ds9: ds9_props, ds1a: ds1a_props},
            profile_id)

    def test_filter_datastores_with_empty_datastores(self):
        self.assertIsNone(self._ds_sel._filter_datastores(
            {}, 1024, None, None, None))

    def _create_host_properties(
            self, parent, connection_state='connected', in_maintenace=False):
        return mock.Mock(connectionState=connection_state,
                         inMaintenanceMode=in_maintenace,
                         parent=parent)

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_get_host_properties')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_get_resource_pool')
    def test_select_best_datastore(self, get_resource_pool, get_host_props):
        host1 = self._create_host('host-1')
        host2 = self._create_host('host-2')
        host3 = self._create_host('host-3')

        host_mounts1 = [mock.Mock(key=host1,
                                  mountInfo=mock.sentinel.ds1_mount_info1),
                        mock.Mock(key=host2,
                                  mountInfo=mock.sentinel.ds1_mount_info2),
                        mock.Mock(key=host3,
                                  mountInfo=mock.sentinel.ds1_mount_info3)]
        host_mounts2 = [mock.Mock(key=host2,
                                  mountInfo=mock.sentinel.ds2_mount_info2),
                        mock.Mock(key=host3,
                                  mountInfo=mock.sentinel.ds2_mount_info3)]
        host_mounts3 = [mock.Mock(key=host1,
                                  mountInfo=mock.sentinel.ds3_mount_info1),
                        mock.Mock(key=host2,
                                  mountInfo=mock.sentinel.ds3_mount_info2)]
        host_mounts4 = [mock.Mock(key=host1,
                                  mountInfo=mock.sentinel.ds4_mount_info1)]

        ds1 = self._create_datastore('ds-1')
        ds1_props = {'summary': self._create_summary(ds1),
                     'host': host_mounts1}

        ds2 = self._create_datastore('ds-2')
        ds2_props = {
            'summary': self._create_summary(
                ds2, free_space=1024, capacity=2048),
            'host': host_mounts2}

        ds3 = self._create_datastore('ds-3')
        ds3_props = {
            'summary': self._create_summary(
                ds3, free_space=512, capacity=2048),
            'host': host_mounts3}

        ds4 = self._create_datastore('ds-3')
        ds4_props = {'summary': self._create_summary(ds4),
                     'host': host_mounts4}

        cluster_ref = mock.sentinel.cluster_ref

        def mock_get_host_properties(host_ref):
            self.assertIsNot(host1, host_ref)
            if host_ref == host2:
                in_maintenance = False
            else:
                in_maintenance = True
            runtime = mock.Mock(spec=['connectionState', 'inMaintenanceMode'])
            runtime.connectionState = 'connected'
            runtime.inMaintenanceMode = in_maintenance
            return {'parent': cluster_ref, 'runtime': runtime}

        get_host_props.side_effect = mock_get_host_properties

        def mock_is_usable(mount_info):
            if (mount_info == mock.sentinel.ds1_mount_info2 or
                    mount_info == mock.sentinel.ds2_mount_info2):
                return False
            else:
                return True

        self._vops._is_usable.side_effect = mock_is_usable

        rp = mock.sentinel.resource_pool
        get_resource_pool.return_value = rp

        # ds1 is mounted to 3 hosts: host1, host2 and host3; host1 is
        # not a valid host, ds1 is not usable in host1, and host3 is
        # in maintenance mode.
        # ds2 and ds3 are mounted to same hosts, and ds2 has a low space
        # utilization. But ds2 is not usable in host2, and host3 is in
        # maintenance mode. Therefore, ds3 and host2 will be selected.
        datastores = {ds1: ds1_props,
                      ds2: ds2_props,
                      ds3: ds3_props,
                      ds4: ds4_props}
        ret = self._ds_sel._select_best_datastore(
            datastores, valid_host_refs=[host2, host3])

        self.assertEqual((host2, rp, ds3_props['summary']), ret)
        self.assertItemsEqual([mock.call(mock.sentinel.ds1_mount_info2),
                               mock.call(mock.sentinel.ds1_mount_info3),
                               mock.call(mock.sentinel.ds2_mount_info2),
                               mock.call(mock.sentinel.ds2_mount_info3),
                               mock.call(mock.sentinel.ds3_mount_info2)],
                              self._vops._is_usable.call_args_list)
        self.assertEqual([mock.call(host3), mock.call(host2)],
                         get_host_props.call_args_list)
        get_resource_pool.assert_called_once_with(cluster_ref)

    def test_select_best_datastore_with_empty_datastores(self):
        self.assertIsNone(self._ds_sel._select_best_datastore({}))

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                'get_profile_id')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_get_datastores')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_filter_datastores')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_select_best_datastore')
    def test_select_datastore(
            self, select_best_datastore, filter_datastores, get_datastores,
            get_profile_id):

        profile_id = mock.sentinel.profile_id
        get_profile_id.return_value = profile_id

        datastores = mock.sentinel.datastores
        get_datastores.return_value = datastores

        filtered_datastores = mock.sentinel.filtered_datastores
        filter_datastores.return_value = filtered_datastores

        best_datastore = mock.sentinel.best_datastore
        select_best_datastore.return_value = best_datastore

        size_bytes = 1024
        req = {self._ds_sel.SIZE_BYTES: size_bytes}
        aff_ds_types = [ds_sel.DatastoreType.VMFS]
        req[ds_sel.DatastoreSelector.HARD_AFFINITY_DS_TYPE] = aff_ds_types
        anti_affinity_ds = [mock.sentinel.ds]
        req[ds_sel.DatastoreSelector.HARD_ANTI_AFFINITY_DS] = anti_affinity_ds
        profile_name = mock.sentinel.profile_name
        req[ds_sel.DatastoreSelector.PROFILE_NAME] = profile_name

        hosts = mock.sentinel.hosts
        self.assertEqual(best_datastore,
                         self._ds_sel.select_datastore(req, hosts))
        get_datastores.assert_called_once_with()
        filter_datastores.assert_called_once_with(
            datastores, size_bytes, profile_id, anti_affinity_ds, aff_ds_types,
            valid_host_refs=hosts)
        select_best_datastore.assert_called_once_with(filtered_datastores,
                                                      valid_host_refs=hosts)

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                'get_profile_id')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_filter_by_profile')
    def test_is_datastore_compliant_true(
            self, filter_by_profile, get_profile_id):
        profile_name = mock.sentinel.profile_name
        datastore = mock.sentinel.datastore

        profile_id = mock.sentinel.profile_id
        get_profile_id.return_value = profile_id
        filter_by_profile.return_value = {datastore: None}

        self.assertTrue(self._ds_sel.is_datastore_compliant(datastore,
                                                            profile_name))
        get_profile_id.assert_called_once_with(profile_name)
        filter_by_profile.assert_called_once_with({datastore: None},
                                                  profile_id)

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                'get_profile_id')
    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                '_filter_by_profile')
    def test_is_datastore_compliant_false(
            self, filter_by_profile, get_profile_id):
        profile_name = mock.sentinel.profile_name
        datastore = mock.sentinel.datastore

        profile_id = mock.sentinel.profile_id
        get_profile_id.return_value = profile_id
        filter_by_profile.return_value = {}

        self.assertFalse(self._ds_sel.is_datastore_compliant(datastore,
                                                             profile_name))
        get_profile_id.assert_called_once_with(profile_name)
        filter_by_profile.assert_called_once_with({datastore: None},
                                                  profile_id)

    def test_is_datastore_compliant_with_empty_profile(self):
        self.assertTrue(self._ds_sel.is_datastore_compliant(
            mock.sentinel.datastore, None))

    @mock.patch('cinder.volume.drivers.vmware.datastore.DatastoreSelector.'
                'get_profile_id')
    def test_is_datastore_compliant_with_invalid_profile(self, get_profile_id):
        profile_name = mock.sentinel.profile_name
        get_profile_id.side_effect = vmdk_exceptions.ProfileNotFoundException
        self.assertRaises(vmdk_exceptions.ProfileNotFoundException,
                          self._ds_sel.is_datastore_compliant,
                          mock.sentinel.datastore,
                          profile_name)
        get_profile_id.assert_called_once_with(profile_name)
